"""
*vula* common functions.
"""
from __future__ import annotations

import copy
import json
import os
import pdb
import re
from base64 import b64decode, b64encode
from ipaddress import ip_address, ip_network
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, Dict, Optional

import click
import pydbus
import yaml
from schema import And, Or, Schema, SchemaError, Use

from .constants import _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
from .notclick import DualUse, Exit  # noqa: F401

try:
    import pygments
    from pygments import formatters, lexers  # noqa: F401
except ImportError:
    pygments = None

bp = pdb.set_trace

memoize = lambda f: (
    lambda d={}: lambda *a: d.setdefault(a, a in d or f(*a))
)()


def chown_like_dir_if_root(path):
    """
    This chowns a file to have the same owner as its parent, if we are running
    as root. Otherwise, it does nothing.

    Normally vula is not run as root. The purpose of this function is to ensure
    that files remain owned by their correct user (which is hopefully the owner
    of the directory they reside in) if/when vula *is* run is root.

    As of this writing, this function is only used by the organize daemon when
    writing the hosts file.

    FIXME: this should probably be used when writing the state file as well?
    """
    if os.getuid() == 0:
        dirstat = os.stat(os.path.dirname(os.path.realpath(path)))
        os.chown(path, dirstat.st_uid, dirstat.st_gid)


def _safer_load(
    yaml_file: str,
    schema: Schema,
    size_constraint_min: int = 0,
    size_constraint_max: int = 65000,
) -> Optional[Dict]:
    """
    *_safer_load* loads opens *yaml_file* and validates it based on a defined
    *schema* and ensures it is not smaller than *size_constraint_min* and not
    larger than *size_constraint_max*.
    """
    log: Logger = getLogger()
    try:
        with open(yaml_file, "r") as file_obj:
            f_size = os.stat(file_obj.fileno()).st_size
            if size_constraint_min or size_constraint_max:
                if (
                    f_size < size_constraint_min
                    or f_size > size_constraint_max
                ):
                    log.info(
                        "File size invalid for constraint: %i bytes for %s",
                        f_size,
                        yaml_file,
                    )
                    return None
            yaml_buf = file_obj.read(size_constraint_max)
    except FileNotFoundError:
        return None
    data: Dict[Any, Any] = yaml.safe_load(yaml_buf)
    try:
        reserialized: Dict[Any, Any] = yaml.safe_load(yaml.safe_dump(data))
        if reserialized == data:
            validated_data: Dict[Any, Any] = schema.validate(data)
        else:
            log.info("Data does not pack and unpack as expected")
            return None
    except SchemaError:
        log.info(
            "Data in %s did not validate against schema: %s", yaml_file, schema
        )
        return None
    return validated_data


class attrdict(dict):
    """
    Dictionary which provides attribute access to its keys.

    For best results, use the schemattrdict subclass rather than using this
    directly.
    """

    def __getattr__(self, key):
        """
        get value of key

        >>> a = attrdict({1:2})
        >>> a.__getattr__(1)
        2

        >>> a.__getattr__(2)
        Traceback (most recent call last):
            ...
        AttributeError: 'attrdict' object has no attribute 2
        """
        if key in self:
            return self[key]
        else:
            raise AttributeError(
                "%r object has no attribute %r" % (type(self).__name__, key)
            )

    def __setattr__(self, key, value):
        if key in self:
            raise ValueError(
                "Programmer error: attempt to set an attribute (%s=%r) of an"
                "instance of %r ( which is an attrdict, which provides "
                "read-only access to keys through the attribute interface). "
                "Attributes which are dictionary keys are not allowed to be "
                "set through the attrdict attribute interface. "
                % (key, value, type(self))
            )
        super(attrdict, self).__setattr__(key, value)


class ro_dict(dict):
    """
    This is a dictionary which is not easy to accidentally update.

    It's not a strong read-only protection, as its values may be mutable and
    the dict itself can actually be updated by using the dict type's update
    method instead of using the object's own exception-raising update method
    definedhere.

    We use this instead of the frozendict from PyPI so that we can be an actual
    dict and thus it is trivially compatible with out other dict mixin classes,
    and because we don't need the feature frozendict provides which we don't
    (hashability). Perhaps in the future we'll decide to use frozendict
    instead.

    >>> ro = ro_dict({'a':1,'b':2,'c':3})
    >>> ro['a'] = 2
    Traceback (most recent call last):
         ...
    ValueError: Attempt to set key 'a' in read-only dictionary

    >>> ro = ro_dict({'a':1})
    >>> ro.update({'a':2})
    Traceback (most recent call last):
         ...
    ValueError: Attempt to update read-only dictionary (...)

    """

    def __setitem__(self, key, value):
        """
        raises a ValueError exception when trying to set a key in a ro_dict

        test adding a non-existing key to ro_dict:
        >>> test_ro_dict = ro_dict({"key": "value"})
        >>> test_ro_dict["different_key"] = "other_value"
        Traceback (most recent call last):
        ...
        ValueError: Attempt to set key 'different_key' in read-only dictionary

        test setting an existing key:
        >>> test_ro_dict = ro_dict({"key": "value"})
        >>> test_ro_dict["key"] = "other_value"
        Traceback (most recent call last):
        ...
        ValueError: Attempt to set key 'key' in read-only dictionary
        """
        raise ValueError(
            "Attempt to set key %r in read-only dictionary" % (key,)
        )

    def update(self, *a, **kw):
        """
        raises a ValueError exception when trying to change a ro_dict using the
        update() method

        test updating an existing key:
        >>> test_ro_dict = ro_dict({"key": "value"})
        >>> test_ro_dict.update({"key": "other_value"})
        Traceback (most recent call last):
        ...
        ValueError: Attempt to update read-only dictionary (*({'key': \
'other_value'},), **{})

        test updating a non-existing key:
        >>> test_ro_dict = ro_dict({"key": "value"})
        >>> test_ro_dict.update({"different_key": "other_value"})
        Traceback (most recent call last):
        ...
        ValueError: Attempt to update read-only dictionary \
(*({'different_key': 'other_value'},), **{})
        """
        raise ValueError(
            "Attempt to update read-only dictionary (*%s, **%s)" % (a, kw)
        )


def raw(value):
    """
    Recursively coerces objects into something serializable.

    This tries to return the original unmodified object whenever possible, to
    avoid precluding the possibility of having automatic YAML object references
    when a sub-object occurs multiple times within a parent object.

    >>> raw(3)
    3
    >>> raw("vula")
    'vula'
    >>> raw(True)
    True
    >>> raw(2.8)
    2.8

    >>> intBool = IntBool(1)
    >>> raw(intBool)
    True

    >>> raw(["one", "two", "three"])
    ['one', 'two', 'three']
    >>> raw(("one", "two", "three"))
    ['one', 'two', 'three']
    >>> sorted(raw({"one", "two", "three"}))
    ['one', 'three', 'two']

    >>> raw({"number" : 1, "letter" : "a"})
    {'number': 1, 'letter': 'a'}

    """
    if type(value) in (int, str, bool, float):
        return value
    if isinstance(value, IntBool):
        return bool(value)
    if hasattr(value, '_dict'):
        return value._dict()
    if type(value) in (list, set, tuple):
        new = list(raw(item) for item in value)
        return (
            value
            if type(value) is list and all(a is b for a, b in zip(new, value))
            else new
        )
    if isinstance(value, dict):
        new = {raw(k): raw(v) for k, v in value.items()}
        return (
            value
            if type(value) is dict
            and all(
                k is k_ and v is v_
                for ((k, v), (k_, v_)) in zip(new.items(), value.items())
            )
            else new
        )
    return str(value)


class serializable(dict):
    def _dict(self):
        """Return serializable as a dictionary

        >>> serializable({1:serializable({2:3})})._dict()
        {1: {2: 3}}

        >>> serializable({1:{2:(3,4,5)}})._dict()
        {1: {2: [3, 4, 5]}}
        """
        return {raw(k): raw(v) for k, v in self.items()}


class schemadict(ro_dict, serializable):

    schema = NotImplemented
    default = None

    def __init__(self, *a, **kw):
        self._as_dict = None
        data = copy.deepcopy(self.default) or {}
        kw = {k: v for k, v in kw.items() if v is not None}
        data.update(*a, **kw)
        assert type(data) == dict
        super(schemadict, self).__init__(self.schema.validate(data))

    def __deepcopy__(self, memo):
        return type(self)(copy.deepcopy(dict(self)))

    def copy(self):
        return self.__deepcopy__(None)

    def _dict(self):
        if self._as_dict is None:
            self._as_dict = super(schemadict, self)._dict()
        return self._as_dict


class schemattrdict(attrdict, schemadict):
    pass


class yamlfile(serializable):
    def write_yaml_file(self, path, mode=None, autochown=False):
        if mode:
            Path(path).touch(mode=mode)

        with click.open_file(
            path, mode='w', encoding='utf-8', atomic=True
        ) as fh:
            fh.write(
                yaml.safe_dump(self._dict(), default_style='', sort_keys=False)
            )
        if autochown:
            chown_like_dir_if_root(path)

    @classmethod
    def from_yaml_file(cls, path):
        with click.open_file(path, mode='r', encoding='utf-8') as fh:
            return cls(yaml.safe_load(fh))


class yamlrepr(serializable):
    r"""
    Function to return a YAML representation of a Serializable in human
    readable format
    >>> myObj = serializable(("ab", "cd"))
    >>> myYaml = yamlrepr(myObj)
    >>> myYaml.__repr__()
    'a: b\nc: d\n'
    """

    def __repr__(self):
        """
        Function to return a YAML representation of a Serializable in human
        readable format
        >>> yamlrepr(serializable({2:3}))
        2: 3
        <BLANKLINE>
        """
        return yaml.safe_dump(self._dict(), default_style='', sort_keys=False)

    @classmethod
    def from_yaml(cls, yaml_str):
        return cls(yaml.safe_load(yaml_str))


class jsonrepr_pp(serializable):
    def __repr__(self):
        r"""
        Function to return a json representation of a Serializable in human
        readable format
        >>> myObj = serializable(("ab", "cd"))
        >>> myJson = jsonrepr_pp(myObj)
        >>> myJson.__repr__()
        '{\n    "a": "b",\n    "c": "d"\n}'
        """
        return json.dumps(self._dict(), indent=4)


class jsonrepr(serializable):
    def __repr__(self):
        """
        Function to return a json representation of a Serializable.
        >>> mySerializable = serializable(("ab", "cd"))
        >>> myJsonRepr = jsonrepr(mySerializable)
        >>> myJsonRepr.__repr__()
        '{"a": "b", "c": "d"}'
        """
        return json.dumps(self._dict())


if pygments is not None:

    #    from pygments.style import Style
    #    from pygments.token import ( Keyword, Name, Comment, String,
    #        Error, Number, Operator, Generic,)
    #
    #    class MyStyle(Style):
    #        default_style = ""
    #        styles = {
    #            Name: 'bold #0f0',
    #        }

    class yamlrepr_hl(yamlrepr):
        def __repr__(self):

            # aaa = list(
            #    pygments.lexers.YamlLexer().get_tokens_unprocessed(
            #        yaml.dump(self._dict(), default_style='', sort_keys=False)
            #    )
            # )

            res = pygments.highlight(
                yaml.safe_dump(
                    self._dict(), default_style='', sort_keys=False
                ),
                pygments.lexers.YamlLexer(),
                pygments.formatters.TerminalTrueColorFormatter(
                    #    style=MyStyle,
                    style='paraiso-dark'
                ),
            )
            return res

    class jsonrepr_hl(jsonrepr):
        def __repr__(self):
            r"""
            Function to return a raw json formatted content with syntax
            highlighting.
            >>> serializableObj = serializable(("ab", "cd"))
            >>> jsonReprObj = jsonrepr(serializableObj)
            >>> jsonReprhlObj = jsonrepr_hl(jsonReprObj)
            >>> jsonReprhlObj.__repr__()
            '\x1b[38;2;231;233;219m{\x1b[39m\n\x1b[38;2;231;233;219m  \x1b[39m\x1b[38;2;91;196;191m"a"\x1b[39m\x1b[38;2;231;233;219m:\x1b[39m\x1b[38;2;231;233;219m \x1b[39m\x1b[38;2;72;182;133m"b"\x1b[39m\x1b[38;2;231;233;219m,\x1b[39m\n\x1b[38;2;231;233;219m  \x1b[39m\x1b[38;2;91;196;191m"c"\x1b[39m\x1b[38;2;231;233;219m:\x1b[39m\x1b[38;2;231;233;219m \x1b[39m\x1b[38;2;72;182;133m"d"\x1b[39m\n\x1b[38;2;231;233;219m}\x1b[39m\n'
            """  # noqa: E501

            res = pygments.highlight(
                json.dumps(self._dict(), indent=2),
                pygments.lexers.JsonLexer(),
                pygments.formatters.TerminalTrueColorFormatter(
                    style='paraiso-dark'
                ),
            )
            return res

else:
    yamlrepr_hl = yamlrepr  # type: ignore
    jsonrepr_hl = jsonrepr  # type: ignore


class Bug(Exception):
    pass


class ConsistencyError(Exception):
    pass


class comma_separated_IPs(object):
    def __init__(self, _str):
        """
        create a list of IP address

        >>> comma_separated_IPs("127.0.0.1,15.15.15.15")
        <comma_separated_IPs('127.0.0.1,15.15.15.15')>
        >>> comma_separated_IPs("fe80::1,fe80::2")
        <comma_separated_IPs('fe80::1,fe80::2')>
        >>> comma_separated_IPs("::1")
        <comma_separated_IPs('::1')>
        >>> comma_separated_IPs("invalid")
        Traceback (most recent call last):
        ...
        ValueError: 'invalid' does not appear to be an IPv4 or IPv6 address
        """
        self._str = str(_str)
        self._items = tuple(
            ip_address(ip) for ip in self._str.split(',') if ip
        )

    def __iter__(self):
        """
        Iterate over ips

        >>> [ip for ip in comma_separated_IPs("::1,192.168.1.1")]
        [IPv6Address('::1'), IPv4Address('192.168.1.1')]
        """
        return iter(self._items)

    def __getitem__(self, idx):
        """
        get item at index

        :param idx: int
        :return: IPv4Address

        >>> test_list = comma_separated_IPs("127.0.0.2,127.0.0.3,127.0.0.4")
        >>> test_list.__getitem__(1)
        IPv4Address('127.0.0.3')


        :param idx: int
        :return: IPv6Address

        >>> test_list = comma_separated_IPs("fe80::1,fe80::2")
        >>> test_list.__getitem__(0)
        IPv6Address('fe80::1')

        """
        return list(self)[idx]

    def __repr__(self):
        """
        Return the type of the class with his IP

        >>> ip = comma_separated_IPs('192.168.0.1')
        >>> ip.__repr__()
        "<comma_separated_IPs('192.168.0.1')>"

        >>> ip = comma_separated_IPs('fe80::1')
        >>> ip.__repr__()
        "<comma_separated_IPs('fe80::1')>"

        """

        """
        get IP

        >>> ips = comma_separated_IPs("192.168.3.2,192.168.0.3,192.168.3.4,
        ... 192.168.0.2,192.168.0.3,192.168.3.4")
        >>> ips.__repr__()
        "<comma_separated_IPs('192.168.3.2,192.168.0.3,192.168.3.4, \
        192.168.0.2,192.168.0.3,192.168.3.4')>"
        """
        return "<%s(%r)>" % (type(self).__name__, self._str)

    def __str__(self):
        """
        Return the comma separated IP addresses as a string

        >>> ip = comma_separated_IPs('fe80::1')
        >>> ip.__str__()
        'fe80::1'

        >>> ip = comma_separated_IPs('192.168.0.1')
        >>> ip.__str__()
        '192.168.0.1'

        >>> ip_multiple = \
        comma_separated_IPs('192.168.29.32,127.0.0.1,149.132.22.70')
        >>> ip_multiple.__str__()
        '192.168.29.32,127.0.0.1,149.132.22.70'

        >>> ip_multiple = \
        comma_separated_IPs('fe80::1,fe80::2,fe80::3')
        >>> ip_multiple.__str__()
        'fe80::1,fe80::2,fe80::3'
        """

        return self._str


class comma_separated_Nets(comma_separated_IPs):
    """
    Instantiate a list of networks

    >>> comma_separated_Nets("blabla")
    Traceback (most recent call last):
       ...
    ValueError: 'blabla' does not appear to be an IPv4 or IPv6 network
    >>> comma_separated_Nets("192.168.0.0/28,192.168.0.1/28")
    Traceback (most recent call last):
       ...
    ValueError: 192.168.0.1/28 has host bits set

    >>> comma_separated_Nets("fe80::1/10,fe80::/10")
    Traceback (most recent call last):
       ...
    ValueError: fe80::1/10 has host bits set

    >>> comma_separated_Nets("192.168.0.0/28,192.168.1.0/25")
    <comma_separated_Nets('192.168.0.0/28,192.168.1.0/25')>

    >>> comma_separated_Nets("fe80::/10,fe80::/10")
    <comma_separated_Nets('fe80::/10,fe80::/10')>
    """

    def __init__(self, _str):
        self._str = str(_str)
        self._items = tuple(
            ip_network(ip) for ip in self._str.split(',') if ip
        )


class constraint(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        """
        Function to return constraint object.
        >>> constraint({4:3})
        constraint({4: 3})
        """
        # FIXME: add kwargs to repr
        return "%s(%s)" % (
            type(self).__name__,
            ", ".join(map(repr, self.args)),
        )

    def validate(self, value):
        """
        >>> constraint.validate(int_range(6,9), 7)
        7
        >>> constraint.validate(int_range(6, 9), 5)
        Traceback (most recent call last):
        ValueError: int_range(6, 9) check failed on False
        """
        # import pdb; pdb.set_trace()
        value = self.constraint(value, *self.args, **self.kwargs)
        if value is not False:
            return value
        else:
            raise ValueError("%s check failed on %r" % (self, value))


class Length(constraint):
    """
    Constraint to check if a string length is equal to, min or max a given
    value.
    """

    @staticmethod
    def constraint(value, length=None, min=None, max=None):
        """
        Validates min and max length of a given string.
        Returns the value of constraints are met else an error is raised.

        >>> Length.constraint("Test", min = 3, max = 5)
        'Test'
        >>> Length.constraint("Vula is cool", length = 12)
        'Vula is cool'
        >>> Length.constraint("Test", length = 3)
        Traceback (most recent call last):
            ...
        ValueError: length is 4, should be 3: 'Test'
        >>> Length.constraint("Test", max = 3)
        Traceback (most recent call last):
            ...
        ValueError: length is 4, should be <=3: 'Test'
        >>> Length.constraint("Test", min = 5)
        Traceback (most recent call last):
            ...
        ValueError: length is 4, should be >=5: 'Test'
        """
        if length is not None and len(value) != length:
            raise ValueError(
                "length is %s, should be %s: %r" % (len(value), length, value)
            )
        if max is not None and len(value) > max:
            raise ValueError(
                "length is %s, should be <=%s: %r" % (len(value), max, value)
            )
        if min is not None and len(value) < min:
            raise ValueError(
                "length is %s, should be >=%s: %r" % (len(value), min, value)
            )
        return value


class int_range(constraint):
    @staticmethod
    def constraint(value, min, max):
        """
        Validates value, which must be an integer between min and max.
        Returns the value if constraints are met, false otherwise.

        >>> int_range.constraint(5, min = 2, max = 10)
        5
        >>> int_range.constraint(5, min = 7, max = 10)
        False
        >>> int_range.constraint(5, min = 2, max = 4)
        False
        >>> int_range.constraint("abc", min = 2, max = 10)
        Traceback (most recent call last):
        ...
        ValueError: invalid literal for int() with base 10: 'abc'
        >>> int_range.constraint(10, 5, 20)
        10
        >>> int_range.constraint(5, 10, 20)
        False
        >>> int_range.constraint(20, 5, 10)
        False
        >>> int_range.constraint(5, 5, 5)
        5
        >>> int_range.constraint(5, 6, 4)
        False
        """
        return min <= int(value) <= max and int(value)


class colon_hex_bytes(bytes):
    def __str__(self):
        """
        Return the hex representation as a string.

        >>> a = colon_hex_bytes.with_len(3).validate(b'ABC')
        >>> a.__str__()
        '41:42:43'
        """
        return ":".join("%x" % byte for byte in self)

    def __repr__(self):
        """
        Return the canonical string representation of the colon_hex_bytes
        object.

        >>> hex_bytes = colon_hex_bytes.with_len(3).validate(b'ABC')
        >>> hex_bytes.__repr__()
        "'41:42:43'"
        """
        return repr(str(self))

    @classmethod
    def with_len(cls, length):
        """
        >>> a = colon_hex_bytes.with_len(3).validate(b'ABC')
        >>> a
        '41:42:43'
        >>> colon_hex_bytes.with_len(3).validate( \
        b'ABCD') # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        schema.SchemaError:
        >>> b = colon_hex_bytes.with_len(3).validate(str(a))
        >>> a == b
        True
        >>> list(a)
        [65, 66, 67]
        >>> a.decode()
        'ABC'
        >>> c = colon_hex_bytes.with_len(4).validate(b'ABCD')
        >>> c
        '41:42:43:44'
        """
        str_len = length * 3 - 1
        return Or(
            And(
                str,
                Length(str_len),
                Use(lambda s: s.split(':')),
                Length(length),
                Use(lambda s: [int(b, 16) for b in s]),
                Use(cls),
            ),
            And(bytes, Length(length), Use(cls)),
            error=f"{{!r}} is not {length} bytes or a {str_len}-char "
            f"colon-separated hex string "
            f"representing {length} bytes",
        )


MACaddr = colon_hex_bytes.with_len(6)
BSSID = colon_hex_bytes.with_len(6)
ESSID = And(str, Length(min=1, max=32))


class b64_bytes(bytes):
    """
    Bytes subclass which automatically stringifies itself to base64, and has a
    repr which shows the first 6 bytes of its base64 encoding.
    """

    def __str__(self):
        """
        Function to return a string representation of entered bytes.

        >>> string = "Test"
        >>> arr = bytearray(string, 'utf-8')
        >>> test = b64_bytes(arr)
        >>> str(test)
        'VGVzdA=='
        >>> b64decode(str(test))
        b'Test'
        """
        return b64encode(self).decode()

    def __repr__(self):
        """
        Function to return a base64 representation of entered bytes

        >>> string = "Vula is cool."
        >>> arr = bytearray(string, 'utf-8')
        >>> b64_bytes(arr).__repr__()
        '<b64:VnVsYS...(13)>'
        """
        return "<b64:%s...(%s)>" % (str(self)[:6], len(self))

    @classmethod
    def with_len(cls, length):
        """
        Produces a schema object which accepts either bytes or base64-encoded
        strings, and returns b64_bytes objects while enforcing a length
        constraint.

        >>> a = b64encode(b'A'*10).decode()
        >>> a, len(a)
        ('QUFBQUFBQUFBQQ==', 16)
        >>> b = b64_bytes.with_len(10).validate(a)
        >>> b
        <b64:QUFBQU...(10)>
        >>> str(b)
        'QUFBQUFBQUFBQQ=='
        >>> bytes(b)
        b'AAAAAAAAAA'
        >>> c = b64_bytes.with_len(10).validate(bytes(b))
        >>> c == b == bytes(b)
        True
        >>> c == b64_bytes.with_len(10).validate(a)
        True

        this test is complicated to be able to pass both with and without the
        bugfix in schema 0.7.4:
        >>> try:
        ...     b64_bytes.with_len(10).validate('123')
        ... except Exception as ex:
        ...     e = ex
        >>> import schema, packaging.version as pkgv
        >>> assert type(e) is schema.SchemaError
        >>> msg = "'123' is not 10 bytes or a 16-char base64 string "\
        "which decodes to 10 bytes"
        >>> if pkgv.parse(schema.__version__) < pkgv.parse('0.7.3'):
        ...     msg += "\\n{!r} is not 10 bytes or a 16-char "
        ...     msg += "base64 string which decodes to 10 bytes"
        >>> assert e.args == (msg,), (e.args, msg)

        """
        b64_length = 4 * ((length // 3) + bool(length % 3))
        return Or(
            And(
                str,
                Length(b64_length),
                Use(b64decode),
                Use(cls),
                Length(length),
            ),
            And(bytes, Length(length), Use(cls)),
            error="{!r} is not %s bytes or a %s-char base64 string which "
            "decodes to %s bytes" % (length, b64_length, length),
            # name = '%s bytes base64-encoded' % (length,), #when we upgrade to
            # the newer schema lib
        )


class IntBool(int):
    """
    This holds values defined as Flexibool in schemas, which allows bools to be
    specified by users in a variety of ways.

    This class exists so that these values can be identified in the 'raw'
    function, which will convert them to normal bools.
    """


Flexibool = And(
    Or(
        bool,
        And(int, lambda n: n in (0, 1)),
        And(
            str,
            Use(
                lambda v: 1
                if v.lower() in ('true', 'yes', 'on', '1', 'y', 'j', 'ja')
                else 0
                if v.lower() in ('false', 'no', 'off', '0', 'n', 'nein', 'nej')
                else 'error'
            ),
        ),
    ),
    Use(IntBool),
    error="invalid value {!r}; must be one of "
    "<true|false|1|0|on|off|yes|no|ja|nein|nej|y|j|n>",
)


class queryable(dict):
    def limit(self, **kw):
        """
        >>> d = {1:{"enabled":True},2:{"enabled":False}}
        >>> d
        {1: {'enabled': True}, 2: {'enabled': False}}
        >>> q = queryable(d)
        >>> q.limit(enabled=True)
        {1: {'enabled': True}}
        >>> q.limit(enabled=False)
        {2: {'enabled': False}}
        >>> q.limit()
        {1: {'enabled': True}, 2: {'enabled': False}}
        """
        return type(self)(
            (name, item)
            for name, item in self.items()
            if all(
                (
                    (v in item[k])
                    if isinstance(item[k], (dict, list, set))
                    else (v == item[k])
                )
                for k, v in kw.items()
            )
        )

    def limit_attr(self, **kw):
        return type(self)(
            (name, item)
            for name, item in self.items()
            if all(
                (
                    (v in getattr(item, k))
                    if isinstance(getattr(item, k), (dict, list, set))
                    else (v == getattr(item, k))
                )
                for k, v in kw.items()
            )
        )

    def by(self, key):
        res = type(self)()
        for item in self.values():
            value = getattr(item, key)

            if isinstance(value, list):
                for subvalue in value:
                    res.setdefault(subvalue, []).append(item)
            elif isinstance(value, dict):
                for subkey, subvalue in value.items():
                    if subvalue:
                        res.setdefault(subkey, []).append(item)
            else:
                res.setdefault(value, []).append(item)
        return res


class chunkable_values(dict):

    """
    This chunks and unchunks dictionary values.

    So, a long value stored under key k becomes many small keys k00..kNN who's
    values can be concatenated to obtain the original long value.

    This is meant to be used for encoding values too large to fit in a TXT
    record. Therefore the chunk size leaves room for the key size plus a two
    digit chunk number plus one more byte for the equals sign in a ZeroConf TXT
    record.

    This was just written for a quick experiment and should be made more robust
    before actually using it.

    >>> d = chunkable_values(a=1, b='0123456789')
    >>> d.chunk(10)
    {'a': 1, 'b00': '012345', 'b01': '6789'}
    >>> d.chunk(7)
    {'a': 1, 'b00': '012', 'b01': '345', 'b02': '678', 'b03': '9'}
    >>> d.chunk(5).unchunk()
    {'a': 1, 'b': '0123456789'}
    >>> d.chunk(4)
    Traceback (most recent call last):
    AssertionError: no room for data with chunk size 4 and key b
    >>> d.chunk(9).chunk(8).unchunk().unchunk() == d
    True
    """

    def chunk(self, length):
        res = {}
        for k, v in list(self.items()):
            if len(k) + len(str(v)) + 1 <= length:
                res[k] = v
            else:
                c = 0
                cs = length - len(k) - 3
                assert (
                    cs >= 1
                ), "no room for data with chunk size %s and key %s" % (
                    length,
                    k,
                )
                while v:
                    res["%s%02d" % (k, c)] = v[:cs]
                    v = v[cs:]
                    c += 1
        return type(self)(res)

    def unchunk(self):
        """Combines chunks of this dictionary.

        >>> chunkable_values({'a01':'23','a00':'01','a02':'45'}).unchunk()
        {'a': '012345'}
        """
        res = {}
        for k, v in list(sorted(self.items())):
            rk = k[:-2]
            try:
                int(k[-2:])
            except ValueError:
                # Expects a ValueError: invalid literal for int() ...  if the
                # last two digits are not a number it is not a chunked key
                res[k] = v
                continue
            res[rk] = res.get(rk, '') + v
        return type(self)(res)


def addrs_in_subnets(addrs, subnets):
    """
    >>> current_subnets={'10.0.0.0/24': ['10.0.0.9', '10.0.0.51',
    ... '10.0.0.17'], '10.0.1.0/24':
    ... ['10.0.1.22', '10.0.1.73'], '10.0.5.0/24': ['10.0.5.21', '10.0.5.63']}
    >>> addrs = ['10.0.0.0/24','10.0.14.0/24', '10.0.5.0/24']
    >>> addrs_in_subnets(addrs,current_subnets)
    ['10.0.0.0/24', '10.0.5.0/24']

    >>> current_subnets={'fe80::/10':['fe80::1', 'fe80::2'],
    ... 'fe80::1:0/10': ['fe80::1:1', 'fe80::1:6' ],
    ... 'fe80::2:0/10': ['fe80::2:1', 'fe80::2:5' ]}
    >>> addrs = ['fe80::/10', 'fe80::ffff:1/10', 'fe80::2:0/10']
    >>> addrs_in_subnets(addrs, current_subnets)
    ['fe80::/10', 'fe80::2:0/10']

    """
    return [
        addr for addr in addrs if any(addr in subnet for subnet in subnets)
    ]


class KeyFile(yamlrepr_hl, schemattrdict, yamlfile):

    schema = Schema(
        {
            "pq_ctidhP512_sec_key": b64_bytes.with_len(74),
            "pq_ctidhP512_pub_key": b64_bytes.with_len(64),
            "vk_Ed25519_sec_key": b64_bytes.with_len(32),
            "vk_Ed25519_pub_key": b64_bytes.with_len(32),
            "wg_Curve25519_sec_key": b64_bytes.with_len(32),
            "wg_Curve25519_pub_key": b64_bytes.with_len(32),
        }
    )


def organize_dbus_if_active():

    """
    Returns a dbus proxy to organize, if it is running. Exits otherwise.

    This is for commands that shouldn't dbus-activate it.
    """

    bus = pydbus.SystemBus()
    if bus.dbus.NameHasOwner(_ORGANIZE_DBUS_NAME):
        return bus.get(_ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH)
    elif _ORGANIZE_DBUS_NAME in bus.dbus.ListActivatableNames():
        raise Exit(
            "Organize is not running (but it is dbus-activatable; use 'vula"
            "start' to start it.)."
        )
    else:
        raise Exit("Organize dbus service is not configured")


def sfmt(
    a,
    base=1000,
    places=0,
    unit='',
    infix='',
    preprefix='',
    prefixes="kMGTPEZYH",
):
    """
    >>> sfmt(0), sfmt(1), sfmt(999)
    ('0', '1', '999')
    >>> sfmt(1e3)
    '1k'
    >>> sfmt(1000)
    '1k'
    >>> sfmt(1499)
    '1k'
    >>> sfmt(1499, places=5)
    '1.49900k'
    >>> sfmt(1500)
    '2k'
    >>> sfmt(1e4)
    '10k'
    >>> sfmt(1_000_000)
    '1M'
    >>> sfmt(10_000_000)
    '10M'
    >>> sfmt(10**27)
    '1H'
    >>> sfmt(10**30)
    '1000H'
    >>> sfmt(1023,base=1024,unit="B",infix="i",prefixes="KMG",preprefix=" ")
    '1023 B'
    >>> sfmt(1024,base=1024,unit="B",infix="i",prefixes="KMG",preprefix=" ")
    '1 KiB'
    >>> sfmt(100_000,base=1024,unit="B",infix="i",prefixes="KMG",preprefix=" ")
    '98 KiB'
    >>> sfmt(1_000_000,base=1024,unit="B",infix="i",prefixes="KMG",
    ... preprefix=" ")
    '977 KiB'
    >>> sfmt(10_000_000, base=1024, unit="B", infix="i", prefixes="KMG",
    ... preprefix=" ", places=1)
    '9.5 MiB'

    """
    i = 0
    while base <= a and i < len(prefixes):
        a /= base
        i += 1
    return (
        ("{:.%sf}{}" % (places if i else 0,))
        .format(
            a,
            preprefix + (prefixes[i - 1].strip() + infix if i else '') + unit,
        )
        .strip()
    )


pprint_bytes = lambda b: sfmt(
    b,
    base=1024,
    unit="B",
    preprefix=" ",
    infix="i",
    prefixes="KMGTPEZYH",
    places=2,
)
format_byte_stats = lambda stats: {
    k: pprint_bytes(v) for k, v in stats.items()
}


def escape_ansi(line: str) -> str:
    """
    Removes ANSI escape sequences from the string.

    >>> escape_ansi("\033[0;31mclean\033[0m")
    'clean'
    >>> escape_ansi("clean")
    'clean'
    """
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub('', line)


if __name__ == "__main__":
    import doctest

    print(doctest.testmod())



import gettext

import pkg_resources

from .qr_code_label import QRCodeLabel

locale_path = pkg_resources.resource_filename('vula', 'locale')
lang_translations = gettext.translation(
    domain="ui.components", localedir=locale_path, fallback=True
)
lang_translations.install()



import json
from unittest.mock import MagicMock, Mock, patch

import vula.verify


class TestVerifyCommands:
    def create_verify_commands(self):
        test_descriptor = {
            '192.168.122.140': {
                'r': '',
                'pk': 'Y' * 43 + '=',
                'c': 'Z' * 86 + '==',
                'addrs': '192.168.122.140',
                'vk': 'A' * 43 + '=',
                'vf': '1638114786',
                'dt': '86400',
                'port': '5354',
                'hostname': 'aula.local.',
                'e': 'False',
                's': 'X' * 86 + '==',
            },
            '192.168.122.150': {
                'r': '',
                'pk': 'G' * 43 + '=',
                'c': 'H' * 86 + '==',
                'addrs': '192.168.122.150',
                'vk': 'A' * 43 + '=',
                'vf': '1638114786',
                'dt': '86400',
                'port': '5354',
                'hostname': 'aula.local.',
                'e': 'False',
                's': 'X' * 86 + '==',
            },
        }

        mockdbus = Mock()
        mockpeer = Mock()
        mockpeer.our_latest_descriptors.return_value = json.dumps(
            test_descriptor
        )
        mockdbus.SystemBus.return_value.get.return_value = mockpeer

        mockclickctx = MagicMock()
        mockclickctx.meta.get.return_value.get.return_value = None

        assert (
            mockdbus.SystemBus().get().our_latest_descriptors()
            == json.dumps(test_descriptor)
        )

        with patch("vula.verify.pydbus", mockdbus):
            # the __wrapped__ call gets rid of the click decorators.
            verify = vula.verify.VerifyCommands.__wrapped__(mockclickctx)

        return verify

    def test_init_loads_descriptors(self):
        verify = self.create_verify_commands()
        assert verify.vk is not None
        assert verify.my_descriptors is not None

    def test_my_vk(self, capsys):
        verify = self.create_verify_commands()

        verify.my_vk()
        output = capsys.readouterr().out
        assert 'A' * 43 + '=' in output
        assert 'Y' * 43 + '=' not in output
        assert 'X' * 86 + '=' not in output
        assert 'Z' * 86 + '=' not in output

    def test_my_descriptor(self, capsys):
        verify = self.create_verify_commands()
        verify.my_descriptor()
        output = capsys.readouterr().out
        assert 'Descriptor for 192.168.122.140:' in output
        assert 'Descriptor for 192.168.122.150' in output



import os  # noqa: F401
import sys
from logging import DEBUG, INFO, WARN, basicConfig, getLogger  # noqa: F401

import click
import pydbus

from . import (  # noqa: F40
    common,
    configure,
    discover,
    discover_alt,
    engine,
    organize,
    peer,
    prefs,
    publish,
    publish_alt,
    status,
    tray,
    verify,
    wg,
)
from .constants import (
    _DATE_FMT,
    _LOG_FMT,
    _ORGANIZE_DBUS_NAME,
    _ORGANIZE_DBUS_PATH,
    _PUBLISH_DBUS_NAME,
    _PUBLISH_DBUS_PATH,
)
from .frontend import ui
from .notclick import Debuggable


@click.version_option()
@click.option(
    "-v",
    "--verbose",
    'log_level',
    flag_value=DEBUG,
    help="Set log level DEBUG",
)
@click.option(
    "-q", "--quiet", 'log_level', flag_value=WARN, help="Set log level WARN"
)
@click.option(
    "-i",
    "--info",
    'log_level',
    flag_value=INFO,
    help="Set log level INFO",
    default=True,
    show_default=True,
)
@click.group(cls=Debuggable, scope=globals(), invoke_without_command=True)
@click.pass_context
def main(ctx, log_level, *args, **kwargs):
    """
    vula tools

    With no arguments, runs "peer show"
    """

    if not log_level:
        log_level = INFO

    basicConfig(
        stream=sys.stdout, level=log_level, datefmt=_DATE_FMT, format=_LOG_FMT
    )

    if ctx.invoked_subcommand is None:
        ctx.invoke(status.main)


@main.command()
@click.option(
    "-q",
    "--quick",
    is_flag=True,
    help="Only wait for services, not configuration",
)
def start(quick):
    "Activate organize daemon via dbus, and report its status"
    bus = pydbus.SystemBus()
    if bus.dbus.NameHasOwner(_ORGANIZE_DBUS_NAME):
        click.echo("start: vula d-bus service is already active")
    else:
        click.echo('start: activating vula organize service via dbus')
        bus.get(_ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH)
        # activate publish here so that it will be done by the time status
        # runs.
        bus.get(_PUBLISH_DBUS_NAME, _PUBLISH_DBUS_PATH)
    status.main(
        args=(('--only-systemd',) if quick else ()), standalone_mode=False
    )


@main.command(short_help="Starts the graphical user interface")
def gui():
    ui.main()


@main.command(short_help="Ensure that system is configured correctly")
@click.option(
    '-n',
    '--dry-run',
    is_flag=True,
    help="Print what would be done, without doing it",
)
def repair(dry_run):
    """
    This checks if the system is configured correctly, and (re)configures it if
    it isn't.
    """
    res = common.organize_dbus_if_active().sync(dry_run)
    if res:
        click.echo("\n".join(res))


@main.command()
def rediscover():
    "Tell organize to ask discover for more peers"
    organize = common.organize_dbus_if_active()  # noqa: F811
    click.echo('Discovering on ' + organize.rediscover())


@main.command()
def release_gateway():
    """
    Stop using vula for the default route.

    This command must be run to roam to a non-vula gateway after using a pinned
    peer as the gateway.

    When the system gateway changes to the IP of a pinned peer, it will
    automatically re-enable that peer as the gateway.
    """
    click.echo(common.organize_dbus_if_active().release_gateway())


for name, value in list(globals().items()):
    if name == 'wg':
        wg = wg.main
        # wg is accessible in debug mode via Debuggable's magic interface
        # but we don't want it in our command list in the GUI because it isn't
        # supported (it is mostly an incomplete replica of the normal wg tools;
        # users should use those instead).
        continue
    cmd = getattr(value, 'main', None)
    if isinstance(cmd, click.Command):
        main.add_command(cmd, name=name)

if __name__ == "__main__":
    main()



from unittest import mock
from unittest.mock import Mock

from nacl.signing import SigningKey

import vula.wg
from vula.common import attrdict


class TestInterface:
    def test_sync_interface(self):
        """
        Try to sync a non existing interface when dryrun is set to True.
        """
        non_existent_interface = vula.wg.Interface("foo")
        key = SigningKey.generate()
        expected_result = [
            '# create interface',
            'ip link add foo type wireguard',
            '# bring up interface',
            'ip link set up foo',
            '# configure interface',
            "WireGuard.set('foo', **{'private_key': '<redacted private key>',"
            " 'listen_port': 9999, 'fwmark': 'foo'})",
        ]

        assert expected_result == non_existent_interface.sync_interface(
            key, 9999, "foo", True
        )

        """
        Try to sync a non existing interface when dryrun is set to False.
        """
        ipr_mock = Mock()
        ipr_mock.link_lookup.side_effect = [[], ['foo']]
        ipr_mock.get_links.return_value = False
        non_existent_interface._ipr = ipr_mock

        wg_mock = Mock()
        wg_mock.set.return_value = True
        non_existent_interface._wg = wg_mock

        query_mock = Mock()
        query_mock.return_value = True
        non_existent_interface.query = query_mock

        expected_result = [
            '# create interface',
            'ip link add foo type wireguard',
            '# bring up interface',
            'ip link set up foo',
            '# configure interface',
            "WireGuard.set('foo', **{'private_key': '<redacted private key>',"
            " 'listen_port': 9999, 'fwmark': 'bar'})",
        ]

        assert expected_result == non_existent_interface.sync_interface(
            key, 9999, "bar", False
        )
        ipr_mock.link.assert_has_calls(
            [
                mock.call('add', ifname='foo', kind='wireguard'),
                mock.call('set', index='foo', state='up'),
            ]
        )
        ipr_mock.get_links.assert_called_with(ifname='foo')

    def test_interface_query(self):
        """
        Ensure that a call to query() clears previously stored data
        """

        interface = vula.wg.Interface("foo")
        interface.update({'test': 'foo'})
        interface.query()
        assert interface == {}

        # Mock the `wg.info` inside the `query` method:
        wg_mock = Mock()
        wg_mock.info.return_value = [
            {'foo': 'bar', 'attrs': {'baz': 'test'}},
            't',
        ]
        interface._wg = wg_mock

        interface.query()

        assert interface == {'baz': 'test', 'peers': []}

        """
        Ensure that a call to query() populates Interface dict accordingly.
        """
        interface = vula.wg.Interface("vula")

        wgdevice_attrs = {
            'attrs': [
                (
                    'WGPEER_A_PUBLIC_KEY',
                    b'hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3J' b'AOLXhFQ=',
                ),
                (
                    'WGPEER_A_PRESHARED_KEY',
                    b'ZGC6L/ZdYjtfwZuipT6rGHZ6AgHhHMTkg' b'bipmyK+Nag=',
                ),
                (
                    'WGPEER_A_LAST_HANDSHAKE_TIME',
                    {
                        'tv_sec': 1671376009,
                        'tv_nsec': 548272819,
                        'latest handshake': 'Sun Dec 18 15:06:49 2022',
                    },
                ),
                ('WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL', 0),
                ('WGPEER_A_TX_BYTES', 220),
                ('WGPEER_A_RX_BYTES', 308),
                ('WGPEER_A_PROTOCOL_VERSION', 1),
                (
                    'WGPEER_A_ENDPOINT',
                    {'family': 2, 'port': 5354, 'addr': '10.89.0.3'},
                ),
                (
                    'WGPEER_A_ALLOWEDIPS',
                    [
                        {
                            'attrs': [
                                ('WGALLOWEDIP_A_CIDR_MASK', 32),
                                ('WGALLOWEDIP_A_FAMILY', 2),
                                ('WGALLOWEDIP_A_IPADDR', '0a:59:00:03'),
                            ],
                            'addr': '10.89.0.3/32',
                        }
                    ],
                ),
            ]
        }

        mock_info_return_value = (
            {
                'cmd': 0,
                'version': 1,
                'reserved': 0,
                'attrs': [
                    ('WGDEVICE_A_LISTEN_PORT', 5354),
                    ('WGDEVICE_A_FWMARK', 555),
                    ('WGDEVICE_A_IFINDEX', 4),
                    ('WGDEVICE_A_IFNAME', 'vula'),
                    (
                        'WGDEVICE_A_PRIVATE_KEY',
                        b'MNQFqcSb6jXU2wbe+U1DIMBuYVTgyIXmUt7TffIWvWw=',
                    ),
                    (
                        'WGDEVICE_A_PUBLIC_KEY',
                        b'R55+QsNFggA7Z9s7iiRtGjkJsJVoj0G3mdimuL33/zQ=',
                    ),
                    ('WGDEVICE_A_PEERS', [wgdevice_attrs]),
                ],
                'header': {
                    'length': 320,
                    'type': 35,
                    'flags': 2,
                    'sequence_number': 255,
                    'pid': 195,
                    'error': None,
                    'target': 'localhost',
                },
            },
        )
        wg = Mock()
        wg.info.return_value = mock_info_return_value
        interface._wg = wg

        expected_result2 = {
            'listen_port': 5354,
            'fwmark': 555,
            'ifindex': 4,
            'ifname': 'vula',
            'private_key': b'MNQFqcSb6jXU2wbe+U1DIMBuYVTgyIXmUt7TffIWvWw=',
            'public_key': b'R55+QsNFggA7Z9s7iiRtGjkJsJVoj0G3mdimuL33/zQ=',
            'peers': [
                {
                    'remove': False,
                    'public_key': 'hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOLX'
                    'hFQ=',
                    'preshared_key': 'ZGC6L/ZdYjtfwZuipT6rGHZ6AgHhHMTkgbipm'
                    'yK+Nag=',
                    'allowed_ips': ['10.89.0.3/32'],
                    'endpoint_addr': '10.89.0.3',
                    'endpoint_port': 5354,
                    'persistent_keepalive': 0,
                    'stats': {
                        'rx_bytes': 308,
                        'tx_bytes': 220,
                        'latest_handshake': 1671376009,
                    },
                }
            ],
        }
        assert interface.query() == expected_result2

    def test_peers_by_pubkey(self):
        interface = vula.wg.Interface("vula")
        interface.update(popluated_interface())

        expected_result = {
            'hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOLXhFQ=': {
                'remove': False,
                'public_key': 'hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOLXhFQ=',
                'preshared_key': 'ZGC6L/ZdYjtfwZuipT6rGHZ6AgHhHMTkgbipmyK+N'
                'ag=',
                'allowed_ips': ['10.89.0.3/32'],
                'endpoint_addr': '10.89.0.3',
                'endpoint_port': 5354,
                'persistent_keepalive': 0,
                'stats': {
                    'rx_bytes': 308,
                    'tx_bytes': 220,
                    'latest_handshake': 1671376009,
                },
            }
        }
        assert interface._peers_by_pubkey == expected_result

    def test_apply_peerconfig(self):
        """
        Prepare a populated Interface with a mocked Wireguard property.
        """
        interface = vula.wg.Interface("vula")
        wg_mock = Mock()
        wg_mock.info.return_value = mock_wg_info_return()
        interface._wg = wg_mock
        existing_peer_pubkey = 'hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOLXhFQ='

        """
        Try to update allowed_ips for an existing peer.
        """

        res = interface.apply_peerconfig(
            attrdict(
                public_key=existing_peer_pubkey,
                allowed_ips=['10.89.0.3/32', '10.89.0.5/32'],
            )
        )

        expected_result = (
            "# allowed_ips is ['10.89.0.3/32']; should be ['10."
            "89.0.3/32', '10.89.0.5/32']\n# reconfigure wiregua"
            "rd peer hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOLXhF"
            "Q=\nvula wg set vula peer hDzSznlwlq9mk07QpNk+Acsf"
            "prrLg2DxSv3JAOLXhFQ= allowed-ips 10.89.0.3/32,10.8"
            "9.0.5/32 "
        )
        assert res == expected_result

        # Verify that Wireguard's `set` method is correctly called to update
        # the existing peer with the new value for allowed_ips.
        wg_mock.set.assert_called_with(
            'vula',
            peer=attrdict(
                public_key='hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOLXhFQ=',
                allowed_ips=['10.89.0.3/32', '10.89.0.5/32'],
            ),
        )

        """
        Try to remove a non-existent peer.
        """
        random_pubkey = "Rbt3m34X1PPIEd/LvW9G0tbImDfcQW0MyvGikM7ayio="
        res = interface.apply_peerconfig(
            attrdict(
                public_key=random_pubkey,
                remove=True,
            )
        )

        assert res == (
            f'# can\'t remove non-existent wireguard peer {random_pubkey}'
        )

        """
        Try to remove an existent peer.
        """

        res = interface.apply_peerconfig(
            attrdict(
                public_key=existing_peer_pubkey,
                remove=True,
            )
        )

        assert res == (
            f'# removing wireguard peer {existing_peer_pubkey}\n# reconfigure '
            f'wireguard peer {existing_peer_pubkey}\nvula wg set vula peer '
            f'{existing_peer_pubkey} remove '
        )

        # Verify that Wireguard's `set` method is correctly called to remove
        # the existing peer.
        wg_mock.set.assert_called_with(
            'vula',
            peer=attrdict(public_key=existing_peer_pubkey, remove=True),
        )


def popluated_interface():
    return {
        'listen_port': 5354,
        'fwmark': 555,
        'ifindex': 4,
        'ifname': 'vula',
        'private_key': b'MNQFqcSb6jXU2wbe+U1DIMBuYVTgyIXmUt7TffIWvWw=',
        'public_key': b'R55+QsNFggA7Z9s7iiRtGjkJsJVoj0G3mdimuL33/zQ=',
        'peers': [
            {
                'remove': False,
                'public_key': 'hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOLXhFQ=',
                'preshared_key': 'ZGC6L/ZdYjtfwZuipT6rGHZ6AgHhHMTkgbipmyK+Nag'
                '=',
                'allowed_ips': ['10.89.0.3/32'],
                'endpoint_addr': '10.89.0.3',
                'endpoint_port': 5354,
                'persistent_keepalive': 0,
                'stats': {
                    'rx_bytes': 308,
                    'tx_bytes': 220,
                    'latest_handshake': 1671376009,
                },
            }
        ],
    }


# Use this simulated wireguard response to populate a vula interface with one
# peer with a public_key of: "hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOLXhFQ="
def mock_wg_info_return():
    return (
        {
            'cmd': 0,
            'version': 1,
            'reserved': 0,
            'attrs': [
                ('WGDEVICE_A_LISTEN_PORT', 5354),
                ('WGDEVICE_A_FWMARK', 555),
                ('WGDEVICE_A_IFINDEX', 4),
                ('WGDEVICE_A_IFNAME', 'vula'),
                (
                    'WGDEVICE_A_PRIVATE_KEY',
                    b'MNQFqcSb6jXU2wbe+U1DIMBuYVTgyIXmUt7TffIWvWw=',
                ),
                (
                    'WGDEVICE_A_PUBLIC_KEY',
                    b'R55+QsNFggA7Z9s7iiRtGjkJsJVoj0G3mdimuL33/zQ=',
                ),
                (
                    'WGDEVICE_A_PEERS',
                    [
                        {
                            'attrs': [
                                (
                                    'WGPEER_A_PUBLIC_KEY',
                                    b'hDzSznlwlq9mk07QpNk+AcsfprrLg2DxSv3JAOL'
                                    b'XhFQ=',
                                ),
                                (
                                    'WGPEER_A_PRESHARED_KEY',
                                    b'ZGC6L/ZdYjtfwZuipT6rGHZ6AgHhHMTkgbipmyK'
                                    b'+Nag=',
                                ),
                                (
                                    'WGPEER_A_LAST_HANDSHAKE_TIME',
                                    {
                                        'tv_sec': 1671376009,
                                        'tv_nsec': 548272819,
                                        'latest handshake': 'Sun Dec 18 15:06:'
                                        '49 2022',
                                    },
                                ),
                                ('WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL', 0),
                                ('WGPEER_A_TX_BYTES', 220),
                                ('WGPEER_A_RX_BYTES', 308),
                                ('WGPEER_A_PROTOCOL_VERSION', 1),
                                (
                                    'WGPEER_A_ENDPOINT',
                                    {
                                        'family': 2,
                                        'port': 5354,
                                        'addr': '10.89.0.3',
                                    },
                                ),
                                (
                                    'WGPEER_A_ALLOWEDIPS',
                                    [
                                        {
                                            'attrs': [
                                                (
                                                    'WGALLOWEDIP_A_CIDR_MASK',
                                                    32,
                                                ),
                                                ('WGALLOWEDIP_A_FAMILY', 2),
                                                (
                                                    'WGALLOWEDIP_A_IPADDR',
                                                    '0a:59:00:03',
                                                ),
                                            ],
                                            'addr': '10.89.0.3/32',
                                        }
                                    ],
                                ),
                            ]
                        }
                    ],
                ),
            ],
            'header': {
                'length': 320,
                'type': 35,
                'flags': 2,
                'sequence_number': 255,
                'pid': 195,
                'error': None,
                'target': 'localhost',
            },
        },
    )



import hashlib
import socket
import threading
import time
import zlib
from logging import Logger, getLogger
from queue import Empty, Queue
from typing import Optional

import click
import nacl.secret
import nacl.utils
import pydbus
from gi.repository import GLib

from .constants import (
    _DISCOVER_ALT_DBUS_NAME,
    _ORGANIZE_DBUS_NAME,
    _ORGANIZE_DBUS_PATH,
)

# assuming 60 byte packets. we have 18 bytes padding/payload.
PAYLOAD_MAX_SIZE = 18


class Discover_Alt:

    dbus = '''
    <node>
      <interface name='local.vula.discoveralt1.Control'>
        <method name='start'>
        </method>
        <method name='stop'>
        </method>
      </interface>
    </node>
    '''

    def __init__(
        self,
        insert_into_vula: bool,
        max_age: int,
        arp_code: int,
        packet_max_length: int,
    ) -> None:
        self.insert_into_vula = insert_into_vula
        self.max_age = max_age
        self.packet_max_length = packet_max_length
        self.log: Logger = getLogger()
        self.socket = socket.socket(
            socket.PF_PACKET,
            socket.SOCK_RAW,
            socket.htons(arp_code),  # maybe filter by packet type
        )
        # self.socket.setblocking(False)
        self.socket.settimeout(1)
        self.active = False
        self.peers_lock = threading.Lock()
        self.peers: dict = {}
        self.sniffing_thread: threading.Thread
        self.packet_queue_lock = threading.Lock()
        self.packet_queue: Queue = Queue()
        self.descriptor_queue_lock = threading.Lock()
        self.descriptor_queue: Queue = Queue()
        self.bus = pydbus.SystemBus()
        self.organize = self.bus.get(_ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH)

    def receive_packet(self) -> bool:
        """
        Listens on the network for ARP packets
        and adds them into the packet queue.
        """
        self.log.debug("receive_packet")
        with self.packet_queue_lock:
            try:
                packet = self.socket.recvfrom(self.packet_max_length)
                self.packet_queue.put(packet)
                return True
            except socket.error:
                pass
        time.sleep(0)  # XXX TODO: do we still need this?
        return False

    def flush_old_peers(self) -> None:
        """
        Flush stale packet streams by clearing out the data structure
        """
        self.log.debug("flush_old_peers")
        t = time.time()
        with self.peers_lock:
            for src_mac in self.peers.keys():
                if t - self.peers[src_mac]["updated"] > self.max_age:
                    self.log.debug("New packet stream detected, clearing data")
                    self.peers[src_mac]["byte_stream"] = b""

    def update_peer(
        self, t: float, src_mac: bytes, candidate: Optional[bytes]
    ) -> None:
        """
        Update the peer by appeding the valid candidate bytes to the byte
        stream and pdating the timestamp.
        """
        self.log.debug("update_peer")
        with self.peers_lock:
            if candidate is None:
                self.peers[src_mac] = {"byte_stream": b"", "updated": t}
                return
            if src_mac not in self.peers:
                self.peers[src_mac] = {"byte_stream": b"", "updated": t}
            self.peers[src_mac]["updated"] = t
            self.log.debug("Received packet")
            self.peers[src_mac]["byte_stream"] += candidate

    def decrypt_all_broadcasts(self) -> None:
        """
        Try to decrypt the packetstream of all entries.
        Add descriptor into queue if succesfull
        """
        self.log.debug("decrypt_all_broadcasts")
        erase_list = []
        with self.peers_lock:
            for src_mac in self.peers.keys():
                descriptor = self.decrypt_broadcast(src_mac)
                if descriptor is not None:
                    self.log.info(
                        f"inserting into descriptor queue {descriptor}"
                    )
                    self.descriptor_queue.put(descriptor)
                    erase_list.append(src_mac)
                else:
                    continue
        now = time.time()
        for src_mac in erase_list:
            self.update_peer(now, src_mac, None)

    def decrypt_broadcast(self, src_mac: bytes) -> Optional[str]:
        """
        Start decryption process
        """
        self.log.debug("decrypt_broadcast")
        return self.decrypt(self.peers[src_mac]["byte_stream"], src_mac)

    def process_descriptors(self) -> bool:
        """
        Empty out descriptor queue by sending them to organize.
        """
        self.log.debug("process_descriptors")
        try:
            descriptor = self.descriptor_queue.get_nowait()
            if self.insert_into_vula:
                self.log.debug(
                    "Inserting descriptor into vula. Descriptor: "
                    + str(descriptor)
                )
                self.organize.process_descriptor_string(descriptor)
                return True
            else:
                self.log.info(descriptor)
                return True
        except Empty:
            return False

    def process_packet(self) -> bool:
        """
        Empty out packet queue and process them
        """
        self.log.debug("process_packet")
        try:
            packet = self.packet_queue.get_nowait()
            candidate = packet[0][42:]
            if candidate[0] != 0 and len(candidate) == PAYLOAD_MAX_SIZE:
                src_mac = packet[0][22:28]
                now = int(time.time())
                self.update_peer(now, src_mac, candidate)
                return True
            else:
                del packet
                del candidate
        except Empty:
            return False
        finally:
            time.sleep(0)
            return False

    def capture(self) -> None:
        """
        Capture thread main body. Captures and processes packets.
        Updates descriptor queue for processing
        """
        self.log.debug("starting capture")
        while self.active:
            self.receive_packet()
            self.log.debug("received packets")
            self.process_packet()
            self.log.debug("processed packets")
            time.sleep(0.5)
        self.log.debug("stopping capture")

    def process(self) -> None:
        """
        Processing thread main body. Processes all packets in queue
        """
        self.log.debug("starting processing")
        while self.active:
            self.decrypt_all_broadcasts()
            self.process_descriptors()
            self.log.debug("processed descriptors")
            time.sleep(0.5)
        self.log.debug("stopping process")

    def flush(self) -> None:
        """
        Flushing thread main body. Clean up datastructure.
        """
        self.log.debug("flush")
        while self.active:
            self.flush_old_peers()
            time.sleep(0.5)
        self.log.debug("stopping flush")

    def decrypt(self, message: bytes, key: bytes) -> Optional[str]:
        """
        Decrypts and decompresses message with given key
        """
        self.log.debug("decrypt")
        key = hashlib.sha256(key).digest()
        box = nacl.secret.SecretBox(key)
        try:
            decrypt = box.decrypt(message.strip(b"\x00"))
            decrypt = zlib.decompress(decrypt)
            if decrypt[10:15] == b"addrs":
                return str(decrypt[10:], "utf-8")
            else:
                return None
        except nacl.exceptions.CryptoError:
            return None

    def start(self) -> None:
        """
        DBus exposed controlling function. Starts the whole process
        """
        self.log.info("Starting alternative discovery")
        self.active = True
        self.sniffing_thread = threading.Thread(target=self.capture)
        self.sniffing_thread.start()
        self.processing_thread = threading.Thread(target=self.process)
        self.processing_thread.start()
        self.flushing_thread = threading.Thread(target=self.flush)
        self.flushing_thread.start()

    def stop(self) -> None:
        """
        DBus exposed controlling function. Stops the whole process
        """
        self.log.info("Stopping alternative discovery")
        click.echo(self.active)
        self.active = False
        click.echo(self.active)

    @classmethod
    def daemon(
        cls,
        insert_into_vula: bool,
        max_age: int,
        arp_code: int,
        packet_max_length: int,
    ) -> None:
        """
        Set up DBus and init object
        """
        loop = GLib.MainLoop()

        discover = cls(insert_into_vula, max_age, arp_code, packet_max_length)

        discover.log.info("dbus enabled")
        system_bus = pydbus.SystemBus()
        system_bus.publish(_DISCOVER_ALT_DBUS_NAME, discover)

        loop.run()


@click.command(short_help="Layer 2 alternate discovery daemon")
@click.option(
    "--insert_into_vula",
    is_flag=True,
    default=True,
    help="Automatically insert into vula.",
)
@click.option(
    "--max_age",
    default=5,
    type=int,
    help="Maximum amount of seconds in between each packet.",
)
@click.option(
    "--arp_code",
    default=0x0806,
    type=int,
    help="EtherType to be used in header",
)
@click.option(
    "--packet_max_length",
    default=60,
    type=int,
    help="Max length of packet to capture",
)
def main(
    insert_into_vula: bool, max_age: int, arp_code: int, packet_max_length: int
) -> None:

    Discover_Alt.daemon(insert_into_vula, max_age, arp_code, packet_max_length)


if __name__ == "__main__":
    main()



import time
from datetime import timedelta
from logging import getLogger
from sys import platform

import click
import pydbus

try:
    from systemd import daemon
except ImportError:
    daemon = None

from click.exceptions import Exit

from .constants import _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
from .notclick import green, red, yellow


@click.command(short_help="Print status")
@click.option(
    '-s',
    '--only-systemd',
    is_flag=True,
    help="Only print systemd service status",
)
def main(only_systemd):
    """
    Print status of systemd services and system configuration
    """
    log = getLogger()

    if not platform.startswith("linux"):
        log.error("so far, the status command only works on linux")
        raise Exit(1)

    def printer(status, service):
        status = "{:^8}".format(status)
        status = (
            green
            if status.strip() in ('active',)
            else yellow
            if status.strip() in ('inactive', 'activatable')
            else red
        )(status)
        click.echo("[{}] {}".format(status, service))

    bus = pydbus.SystemBus()

    if daemon is not None and daemon.booted() == 1:
        systemd = bus.get(".systemd1")
        for service in [
            "publish",
            "discover",
            "organize",
            "organize-monolithic",
        ]:
            unit_name = "vula-%s.service" % (service,)
            try:
                unit = bus.get('.systemd1', systemd.GetUnit(unit_name))
            except Exception as ex:
                if 'NoSuchUnit' in str(ex):
                    if 'monolithic' not in unit_name:
                        # hide NoSuchUnit for monolithic service
                        printer('disabled', unit_name)
                else:
                    printer(str(ex), unit_name)
            else:
                printer(
                    unit.ActiveState,
                    "%-35s (%s)"
                    % (
                        unit_name,
                        timedelta(
                            seconds=int(
                                time.time()
                                - unit.StateChangeTimestamp / 1000000
                            )
                        ),
                    ),
                )

    else:
        printer("inactive", "systemd")

    if bus.dbus.NameHasOwner(_ORGANIZE_DBUS_NAME):
        if only_systemd:
            printer("active", _ORGANIZE_DBUS_NAME + ' dbus service')
        else:
            printer(
                "active",
                _ORGANIZE_DBUS_NAME
                + ' dbus service (running repair --dry-run)',
            )
            organize = bus.get(_ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH)
            sync_todo = organize.sync(True)
            if sync_todo:
                printer(
                    "needs repair",
                    "{} enabled peers; {} disabled".format(
                        str(len(organize.peer_ids('enabled'))),
                        str(len(organize.peer_ids('disabled'))),
                    ),
                )
                for line in sync_todo:
                    printer("needs repair", "[#] %s" % (line,))
            else:
                printer(
                    "active",
                    "{} enabled peers correctly configured; "
                    "{} disabled".format(
                        str(len(organize.peer_ids('enabled'))),
                        str(len(organize.peer_ids('disabled'))),
                    ),
                )

    elif _ORGANIZE_DBUS_NAME in bus.dbus.ListActivatableNames():
        printer("activatable", _ORGANIZE_DBUS_NAME + ' dbus service')
    else:
        printer("not installed", _ORGANIZE_DBUS_NAME + ' dbus service')



from __future__ import annotations

from ipaddress import ip_network

import click
import yaml
from schema import Optional, Schema, Use

from .common import (
    DualUse,
    Flexibool,
    organize_dbus_if_active,
    schemattrdict,
    yamlrepr_hl,
)


class Prefs(yamlrepr_hl, schemattrdict):
    # IPv6 analysis: not ipv6 ready
    # Please enhance this class to support ipv6

    schema = Schema(
        {
            'pin_new_peers': Flexibool,
            'auto_repair': Flexibool,
            'subnets_allowed': [Use(ip_network)],
            'subnets_forbidden': [Use(ip_network)],
            'iface_prefix_allowed': [str],
            'accept_nonlocal': Flexibool,
            'local_domains': [str],
            'ephemeral_mode': Flexibool,
            'accept_default_route': Flexibool,
            Optional('overwrite_unpinned'): Flexibool,  # TODO
            Optional('expire_time'): Use(int),  # TODO
            'record_events': Flexibool,
        }
    )

    default = dict(
        pin_new_peers=False,
        accept_nonlocal=False,
        auto_repair=True,
        subnets_allowed=[
            "10.0.0.0/8",
            "192.168.0.0/16",
            "172.16.0.0/12",
            "169.254.0.0/16",
        ],
        subnets_forbidden=[],
        iface_prefix_allowed=[
            'en',
            'eth',
            'wl',
            'thunderbolt',
            # 'mpqemubr0',
        ],
        local_domains=["local."],
        ephemeral_mode=False,
        accept_default_route=True,
        record_events=False,
        expire_time=3600,
        overwrite_unpinned=True,
    )


@DualUse.object(
    invoke_without_command=True,
)
@click.pass_context
class PrefsCommands(object):
    """
    View and modify preferences
    """

    def __init__(self, ctx):
        self.organize = organize_dbus_if_active()
        if ctx.invoked_subcommand is None:
            click.echo(self.show().strip())

    @DualUse.method()
    def show(self):
        """
        Show preferences
        """
        res = str(yamlrepr_hl(yaml.safe_load(self.organize.show_prefs())))
        return res

    @DualUse.method()
    @click.argument('pref', type=str)
    @click.argument('value', type=str)
    def set(self, pref, value):
        """
        Set a preference to a value
        """
        return self.organize.set_pref(pref, value)

    @DualUse.method()
    @click.argument('pref', type=str)
    @click.argument('value', type=str)
    def add(self, pref, value):
        """
        Merge a preference value into a list or dict
        """
        return self.organize.add_pref(pref, value)

    @DualUse.method()
    @click.argument('pref', type=str)
    @click.argument('value', type=str)
    def remove(self, pref, value):
        """
        Remove a preference value from a list or dict
        """
        return self.organize.remove_pref(pref, value)


main = PrefsCommands.cli



import gettext
import json
import tkinter as tk

from vula.frontend import DataProvider
from vula.frontend.components import QRCodeLabel
from vula.frontend.constants import (
    BACKGROUND_COLOR,
    FONT,
    FONT_SIZE_HEADER,
    FONT_SIZE_TEXT_L,
    IMAGE_BASE_PATH,
    TEXT_COLOR_HEADER,
    TEXT_COLOR_WHITE,
)
from vula.peer import Descriptor

from .popupMessage import PopupMessage

_ = gettext.gettext


class VerificationKeyOverlay(tk.Toplevel):
    data = DataProvider()

    def __init__(self, parent: tk.Tk) -> None:
        tk.Toplevel.__init__(self, parent)
        self.root = parent

        self.wm_transient(self.root)

        # Root Window properties
        self.title("Vula | Verification Key")
        self.geometry("510x510")
        self.configure(bg=BACKGROUND_COLOR)

        # Create the Frames for diffrent sections
        title_frame = tk.Frame(
            self,
            bg=BACKGROUND_COLOR,
            width=510,
            height=70,
            pady=10,
            padx=10,
        )
        text_frame = tk.Frame(
            self, bg=BACKGROUND_COLOR, width=510, height=50, padx=20
        )
        qr_frame = tk.Frame(self, bg=BACKGROUND_COLOR, width=510, height=390)

        # Create the grid for the root TK
        self.grid_rowconfigure(2, weight=1)
        self.grid_columnconfigure(0, weight=1)

        # Place the Frames on the grid
        title_frame.grid(row=0, sticky="nw")
        text_frame.grid(row=1, sticky="nw")
        qr_frame.grid(row=2, sticky="nsew")

        # Add a Title to the Window
        tk.Label(
            title_frame,
            text="Verification Key",
            bg=BACKGROUND_COLOR,
            fg=TEXT_COLOR_HEADER,
            font=(FONT, FONT_SIZE_HEADER),
        ).pack()

        my_descriptors = {
            ip: Descriptor(d)
            for ip, d in json.loads(self.data.our_latest_descriptors()).items()
        }

        text_frame.grid_rowconfigure(1, weight=1)
        text_frame.grid_columnconfigure(2, weight=1)
        self.button_image = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'clipboard.png'
        )

        for d in my_descriptors.values():
            vk = "{}".format(d.vk)

            # Verification Key String Label
            tk.Label(
                text_frame,
                text=vk,
                bg=BACKGROUND_COLOR,
                fg=TEXT_COLOR_WHITE,
                font=(FONT, FONT_SIZE_TEXT_L),
            ).grid(row=0, column=0)

            copy_button = tk.Button(
                text_frame,
                text="Copy",
                image=self.button_image,
                borderwidth=0,
                highlightthickness=0,
                relief="sunken",
                background=BACKGROUND_COLOR,
                activebackground=BACKGROUND_COLOR,
                activeforeground=BACKGROUND_COLOR,
                command=lambda key=vk: self.add_to_clipbaord(key),
            )
            copy_button.grid(row=0, column=1)

            # Verification Key QR Code Image
            qr_data = "local.vula:vk:" + str(vk)
            qr_code = QRCodeLabel(parent=qr_frame, qr_data=qr_data, resize=1)
            qr_code.configure(background=BACKGROUND_COLOR)
            qr_code.pack()

    def add_to_clipbaord(self, text: str) -> None:
        self.clipboard_clear()
        self.clipboard_append(text)
        PopupMessage.showPopupMessage(
            "Information", "Verification key copied to clipboard"
        )



"""
 *vula-organize* is a daemon that reads configuration files as well as peer
 descriptors. It configures the `vula` WireGuard device for new peers. It
 requires administrative privileges for managing the `vula` device and for
 modifying routing information. It does not need network access.

"""
from __future__ import annotations

import os
import pdb
import time
from ipaddress import ip_address, ip_network
from logging import Logger, getLogger
from pathlib import Path
from platform import node

import click
import pydbus
from gi.repository import GLib
from schema import And
from schema import Optional as Optional_
from schema import Schema, Use

from .common import (
    addrs_in_subnets,
    attrdict,
    b64_bytes,
    chown_like_dir_if_root,
    jsonrepr,
    memoize,
    raw,
    schemattrdict,
    yamlrepr,
    yamlrepr_hl,
)
from .configure import Configure
from .constants import (
    _DEFAULT_INTERFACE,
    _DEFAULT_TABLE,
    _DISCOVER_DBUS_NAME,
    _DISCOVER_DBUS_PATH,
    _DOMAIN,
    _FWMARK,
    _IP_RULE_PRIORITY,
    _ORGANIZE_CONF_FILE,
    _ORGANIZE_DBUS_NAME,
    _ORGANIZE_HOSTS_FILE,
    _ORGANIZE_KEYS_CONF_FILE,
    _PUBLISH_DBUS_NAME,
    _PUBLISH_DBUS_PATH,
    _WG_PORT,
    IPv4_GW_ROUTES,
)
from .csidh import ctidh, ctidh_parameters, hkdf
from .discover import Discover
from .engine import Engine, Result
from .notclick import DualUse
from .peer import Descriptor, PeerCommands, Peers
from .prefs import Prefs, PrefsCommands
from .publish import Publish
from .sys import Sys


class SystemState(schemattrdict):

    """
    The SystemState object stores the parts of the system's state which are
    relevant to events in the organize state engine. The object should be
    updated (replaced) whenever these values change, via the
    event_NEW_SYSTEM_STATE event.
    """

    schema = Schema(
        dict(
            current_subnets={Optional_(Use(ip_network)): [Use(ip_address)]},
            our_wg_pk=b64_bytes.with_len(32),
            gateways=[Use(ip_address)],
        )
    )

    default = dict(
        our_wg_pk=b'\x00' * 32,
        current_subnets={},
        gateways=[],
    )

    @classmethod
    def read(cls, organize):
        subnets, gateways = organize.sys._get_system_state()
        return cls(
            gateways=gateways,
            current_subnets=subnets,
            our_wg_pk=organize.our_wg_pk,
        )

    @property
    def current_ips(self):
        """
        Currently assigned IP addresses of the system.

        >>> SystemState().current_ips
        []
        >>> SystemState(current_subnets={'10.0.0.0/24': ['10.0.0.1',
        ... '10.0.0.2']}).current_ips
        [IPv4Address('10.0.0.1'), IPv4Address('10.0.0.2')]
        >>> SystemState(current_subnets={'10.0.0.0/24': ['10.0.0.1'],
        ... '192.168.0.0/24': ['192.168.1.1']}).current_ips
        [IPv4Address('10.0.0.1'), IPv4Address('192.168.1.1')]
        >>> SystemState(current_subnets={'FE80::/10': ['FE80::FFFF:FFFE',
        ... 'FE80::FFFF:FFFD']}).current_ips
        [IPv6Address('fe80::ffff:fffe'), IPv6Address('fe80::ffff:fffd')]
        >>> SystemState(current_subnets={'FE80::/10': ['FE80::FFFF:FFFE'],
        ... 'FC00::/7': ['FC00::FFFF:FFFE']}).current_ips
        [IPv6Address('fe80::ffff:fffe'), IPv6Address('fc00::ffff:fffe')]
        """
        return [
            ip for subnet in self.current_subnets.values() for ip in subnet
        ]


class OrganizeState(Engine, yamlrepr_hl):

    schema = Schema(
        And(
            Use(dict),
            {
                'prefs': Use(Prefs),
                'peers': Use(Peers),
                'system_state': Use(SystemState),
                'event_log': object,
            },
            And(
                lambda state: not state['peers'].conflicts,
                error="conflicting peers: {[peers].conflicts}",
            ),
        )
    )

    default = dict(
        prefs=Prefs.default, system_state={}, peers={}, event_log=[]
    )

    restricted = [
        "peers.*.descriptor"
    ]  # FIXME: implement filter for 1-op direct events?

    def _check_freshness(self, descriptor):
        # FIXME: check dt and vf here
        return True

    def record(self, res):
        if self.prefs.record_events:
            self.event_log.append(raw(res))

    @Engine.event
    def event_VERIFY_AND_PIN_PEER(self, vk, hostname):
        _id = self.peers.with_hostname(hostname).id
        if vk == _id:
            self.action_EDIT('SET', ['peers', vk, 'verified'], True)
            self.action_EDIT('SET', ['peers', vk, 'pinned'], True)
        else:
            raise Exception(f"Expected {hostname}: expected: {vk} have: {_id}")

    @Engine.event
    def event_USER_REMOVE_PEER(self, query):
        peer = self.peers.query(query)
        if peer:
            self.action_REMOVE_PEER(peer)
        else:
            self.action_IGNORE("no such peer")

    @Engine.event
    def event_USER_PEER_ADDR_ADD(self, vk, ip):
        "user added IP address"
        peer = self.peers.get(vk)
        if peer:
            self.action_PEER_ADDR_ADD(peer, vk, ip)
        else:
            self.action_IGNORE("no such peer")

    @Engine.action
    def action_PEER_ADDR_ADD(self, peer, vk, ip):
        "added IP address"
        ipa = ip_address(ip)
        self._SET(('peers', vk, 'IPv%saddrs' % (ipa.version,), ip), True)
        self.result.add_triggers(sync_peer=(peer.id,))

    @Engine.event
    def event_USER_PEER_ADDR_DEL(self, vk, ip):
        "user removed IP address"
        peer = self.peers.get(vk)
        if peer:
            self.action_PEER_ADDR_DEL(peer, vk, ip)
        else:
            self.action_IGNORE("no such peer")

    @Engine.action
    def action_PEER_ADDR_DEL(self, peer, vk, ip):
        "removed IP address"
        ipa = ip_address(ip)
        self._REMOVE(('peers', vk, 'IPv%saddrs' % (ipa.version,)), ip)
        self.result.add_triggers(sync_peer=(peer.id,))
        self.result.add_triggers(
            remove_routes=(ip + ('/32' if ipa.version == 4 else '/128'),)
        )

    @Engine.event
    def event_USER_EDIT(self, operation, path, value):
        self.action_EDIT(operation, path, value)

    @Engine.event
    def event_RELEASE_GATEWAY(self):
        cur_gw = list(self.peers.limit(use_as_gateway=True).values())
        if cur_gw:
            self.action_EDIT(
                'SET', ['peers', cur_gw[0].id, 'use_as_gateway'], False
            )
            self.result.add_triggers(get_new_system_state=())
        else:
            self.action_IGNORE("no current gateway peer")

    @Engine.action
    def action_EDIT(self, operation, path, value):
        getattr(self, '_' + operation)(path, value)
        if path[0] == 'peers':
            # bug: this doesn't work for the REMOVE operation where there is no
            # path[1] but the peer remove command calls event_USER_REMOVE_PEER
            # so that is ok for now
            self.result.add_triggers(sync_peer=(path[1],))
        elif path[0] == 'prefs':
            self.result.add_triggers(get_new_system_state=())
        # FIXME: with more granular actions for peers and prefs, we could
        # remove this call remove_unknown trigger. but for now, it is still
        # necessary.
        self.result.add_triggers(remove_unknown=())

    @Engine.event
    def event_NEW_SYSTEM_STATE(self, new_system_state):
        self.action_ADJUST_TO_NEW_SYSTEM_STATE(new_system_state)

    @Engine.action
    def action_ADJUST_TO_NEW_SYSTEM_STATE(self, new_system_state):
        # IPv6 analysis: not ipv6 ready
        # Please enhance this function to support ipv6
        cur_gw = list(self.peers.limit(use_as_gateway=True).values())
        cur_gw = cur_gw and cur_gw[0]
        if (
            cur_gw
            and (not cur_gw.pinned)
            and (
                not (set(cur_gw.enabled_ips) & set(new_system_state.gateways))
            )
        ):
            # if a non-pinned peer had our gateway IP but no longer does,
            # remove its gateway flag
            self._SET(('peers', cur_gw.id, 'use_as_gateway'), False)
            self.result.add_triggers(remove_routes=(IPv4_GW_ROUTES,))
        if not (cur_gw and cur_gw.pinned):
            # if there isn't a pinned peer acting as the gateway.
            # FIXME: this could set two peers as the gateway if the system has
            # multiple default routes! we need to think about how to handle
            # that...
            for gateway in new_system_state.gateways:
                hit = self.peers.by('enabled_ips').get(gateway)
                if hit:
                    self._SET(('peers', hit[0].id, 'use_as_gateway'), True)
                    self.result.add_triggers(sync_peer=(hit[0].id,))
                    # first hit wins (there could be multiples if we have
                    # multiple default routes, but only one can get
                    # allowedips=/0 so we just take the first one.)
                    break
        for peer in self.peers.limit(pinned=False).values():
            if not addrs_in_subnets(
                peer.enabled_ips, new_system_state.current_subnets
            ):
                # remove unpinned peers that are no longer local
                self.action_REMOVE_PEER(peer)
        self._SET('system_state', new_system_state)

        # TODO:
        # remove endpoints from pinned peers that became non-local

    @Engine.event
    def event_INCOMING_DESCRIPTOR(self, descriptor):

        if str(descriptor.pk) == str(self.system_state.our_wg_pk):
            return self.action_IGNORE(descriptor, "has our wg pk")

        existing_peer = self.peers.get(descriptor.id)

        if not self._check_freshness(descriptor):
            return self.action_REJECT(descriptor, "timestamp too old")

        if existing_peer and descriptor.vf <= existing_peer.descriptor.vf:
            return self.action_IGNORE(descriptor, "replay")

        if not addrs_in_subnets(
            descriptor.addrs, self.system_state.current_subnets
        ):
            return self.action_REJECT(
                descriptor,
                "wrong subnet, current subnets are %r"
                % (self.system_state.current_subnets,),
            )

        if not any(
            descriptor.hostname.endswith(domain)
            for domain in self.prefs.local_domains
        ):
            return self.action_REJECT(descriptor, "invalid domain")

        conflicting_peers = self.peers.conflicts_for_descriptor(descriptor)

        if conflicting_peers:

            for conflicting_peer in conflicting_peers:
                if conflicting_peer.pinned:
                    return self.action_REJECT(
                        descriptor, ("conflict", conflicting_peer)
                    )

            for conflicting_peer in conflicting_peers:
                self.action_REMOVE_PEER(conflicting_peer)

        if existing_peer:
            self.action_UPDATE_PEER_DESCRIPTOR(existing_peer, descriptor)
        else:
            self.action_ACCEPT_NEW_PEER(descriptor)

    @Engine.action
    def action_ACCEPT_NEW_PEER(self, descriptor):
        peer = descriptor.make_peer(
            pinned=False
            if descriptor.ephemeral
            else bool(self.prefs.pin_new_peers)
        )
        self._SET(('peers', descriptor.id), peer)
        if set(descriptor.addrs) & set(self.system_state.gateways):
            # BUG: this will fail (new peer won't be accepted) if the new peer
            # has a gateway IP and there is already another gateway. it fails
            # closed, but should perhaps do so more gracefully? this could only
            # happen if (1) the existing gateway is pinned or, if (2) there are
            # multiple default routes (because otherwise the other gateway
            # would've already been removed before action_ACCEPT_NEW_PEER was
            # called).
            self._SET(('peers', descriptor.id, 'use_as_gateway'), True)
        self.result.add_triggers(sync_peer=(peer.id,))

    @Engine.action
    def action_UPDATE_PEER_DESCRIPTOR(self, peer, desc):
        self._SET(('peers', peer.id, 'descriptor'), desc)
        self._ADD(
            ('peers', peer.id, 'IPv4addrs'),
            {i: True for i in desc.IPv4addrs if i not in peer.IPv4addrs},
        )
        self._ADD(
            ('peers', peer.id, 'IPv6addrs'),
            {i: True for i in desc.IPv6addrs if i not in peer.IPv6addrs},
        )
        if desc.hostname not in peer.nicknames:
            self._SET(('peers', peer.id, 'nicknames', desc.hostname), True)

        if set(desc.addrs) & set(self.system_state.gateways):
            self._SET(('peers', desc.id, 'use_as_gateway'), True)

        self.result.add_triggers(sync_peer=(peer.id,))

    @Engine.action
    def action_REMOVE_PEER(self, peer):
        # IPv6 analysis: not ipv6 ready.
        # Please enhance this function to support ipv6
        self._REMOVE('peers', peer.id)
        self.result.add_triggers(
            remove_wg_peer=(str(peer.wg_pk),),
            remove_routes=(tuple(map(str, peer.routes)),),
        )
        if peer.use_as_gateway:
            # FIXME: this doesn't specify the routing table. maybe need to make
            # triggers api accept kwargs too?
            self.result.add_triggers(remove_routes=(IPv4_GW_ROUTES,))
            # these routes are currently only removed because we still call
            # sync (aka full repair) on system state change

    @Engine.action
    def action_REJECT(self, descriptor, reason):
        pass

    @Engine.action
    def action_IGNORE(self, *a):
        # same as reject, different for logging
        pass

    @Engine.action
    def action_LOG(self, message, *args):
        pass  # the decorator actually does the logging


@DualUse.object(
    short_help="Maintain routes and wg peer configurations",
    invoke_without_command=True,
)
@click.option(
    "-t",
    "--table",
    type=int,
    show_default=True,
    default=_DEFAULT_TABLE,
    help="Which routing table to use",
)
@click.option(
    "-i",
    "--interface",
    show_default=True,
    default=_DEFAULT_INTERFACE,
    help='WireGuard interface name',
)
@click.option(
    "-c",
    "--state-file",
    default=_ORGANIZE_CONF_FILE,
    show_default=True,
    help="YAML state file",
)
@click.option(
    "-k",
    "--keys-file",
    default=_ORGANIZE_KEYS_CONF_FILE,
    show_default=True,
    help="YAML configuration file for cryptographic keys",
)
@click.option(
    "-p",
    "--port",
    default=_WG_PORT,
    show_default=True,
    help="path to base directory for organize state",
)
@click.option(
    "-m",
    "--fwmark",
    default=_FWMARK,
    show_default=True,
    help="path to base directory for organize state",
)
@click.option(
    "-r",
    "--ip-rule-priority",
    default=_IP_RULE_PRIORITY,
    show_default=True,
    help="path to base directory for organize state",
)
@click.pass_context
class Organize(attrdict):
    """
    vula's organize daemon processes events such as new descriptors or system
    configuration changes, and maintains the configuration of the wireguard
    interface and routing table.

    It provides dbus interfaces for client tools to view and modify its state.
    """

    dbus = '''
    <node>
      <interface name='local.vula.organize1.Sync'>
        <method name='sync'>
          <arg type='b' name='dry_run' direction='in'/>
          <arg type='as' name='response' direction='out'/>
        </method>
      </interface>
      <interface name='local.vula.organize1.Debug'>
        <method name='dump_state'>
          <arg type='s' name='response' direction='out'/>
          <arg type='b' name='interactive' direction='in'/>
        </method>
        <method name='test_auth'>
            <arg type='b' name='interactive' direction='in'/>
            <arg type='s' name='response' direction='out'/>
        </method>
      </interface>
      <interface name='local.vula.organize1.Peers'>
        <method name='show_peer'>
          <arg type='s' name='query' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='peer_descriptor'>
          <arg type='s' name='query' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='peer_ids'>
          <arg type='s' name='which' direction='in'/>
          <arg type='as' name='response' direction='out'/>
        </method>
        <method name='rediscover'>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='set_peer'>
          <arg type='s' name='vk' direction='in'/>
          <arg type='as' name='path' direction='in'/>
          <arg type='s' name='value' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='remove_peer'>
          <arg type='s' name='vk' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='peer_addr_add'>
          <arg type='s' name='vk' direction='in'/>
          <arg type='s' name='value' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='peer_addr_del'>
          <arg type='s' name='vk' direction='in'/>
          <arg type='s' name='value' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='our_latest_descriptors'>
          <arg type='s' name='descriptors' direction='out'/>
        </method>
        <method name='get_vk_by_name'>
          <arg type='s' name='hostname' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='verify_and_pin_peer'>
          <arg type='s' name='vk' direction='in'/>
          <arg type='s' name='hostname' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
      </interface>
      <interface name='local.vula.organize1.ProcessDescriptor'>
        <method name='process_descriptor_string'>
          <arg type='s' name='descriptor' direction='in'/>
          <arg type='s' name='response' direction='out'/>
        </method>
      </interface>
      <interface name='local.vula.organize1.Prefs'>
        <method name='show_prefs'>
          <arg type='s' name='response' direction='out'/>
        </method>
        <method name='set_pref'>
          <arg type='s' name='response' direction='out'/>
          <arg type='s' name='pref' direction='in'/>
          <arg type='s' name='value' direction='in'/>
        </method>
        <method name='add_pref'>
          <arg type='s' name='response' direction='out'/>
          <arg type='s' name='pref' direction='in'/>
          <arg type='s' name='value' direction='in'/>
        </method>
        <method name='remove_pref'>
          <arg type='s' name='response' direction='out'/>
          <arg type='s' name='pref' direction='in'/>
          <arg type='s' name='value' direction='in'/>
        </method>
        <method name='release_gateway'>
          <arg type='s' name='response' direction='out'/>
        </method>
      </interface>
    </node>
    '''

    def __init__(self, ctx, **kw):
        self.update(**kw)
        self.log: Logger = getLogger()
        self.log.debug("Debug level logging enabled")
        self._configure = Configure(keys_conf_file=self.keys_file)
        self._ctidh_dh = None
        self._keys = self._configure.generate_or_read_keys()
        self.sys = Sys(self)
        self._state: OrganizeState = self._load_state()
        self._state.trigger_target = self.sys
        self._state.save = self.save
        self._state.debug_log = self.log.debug
        self._latest_descriptors = {}

        if ctx.invoked_subcommand is None:
            self.run(monolithic=False)

    def ctidh_dh(self, pk):
        if self._ctidh_dh is None:
            self.log.debug("Initializing CTIDH")
            _ctidh = ctidh(ctidh_parameters)
            sk = bytes(self._keys.pq_ctidhP512_sec_key)

            @memoize
            def _ctidh_dh(pk: bytes):
                self.log.debug("Generating CTIDH PSK for pk {}".format(pk))
                return _ctidh.dh(
                    _ctidh.private_key_from_bytes(sk),
                    _ctidh.public_key_from_bytes(pk),
                )

            self._ctidh_dh = _ctidh_dh
        raw_key = self._ctidh_dh(bytes(pk))
        # XXX To ensure that even if CTIDH is broken, we should integrate a DH
        # from a Curve-448 using a hybrid construction that is as secure as the
        # most secure of either (ctidh, curve448)
        psk = hkdf(raw_key)
        return psk

    @property
    def our_wg_pk(self):
        return str(self._keys.wg_Curve25519_pub_key)

    def our_latest_descriptors(self):
        return repr(jsonrepr(self._latest_descriptors))

    def _construct_service_descriptor(
        self, ip_addrs: str, vf: int
    ) -> Descriptor:
        self.log.info("Constructing service descriptor id: %s", vf)
        # XXX add Curve448 pk for hybrid DH
        return Descriptor(
            {
                "pk": self._keys.wg_Curve25519_pub_key,
                "c": self._keys.pq_ctidhP512_pub_key,
                "addrs": ip_addrs,
                "vk": self._keys.vk_Ed25519_pub_key,
                "vf": vf,
                "dt": "86400",
                "port": str(self.port),
                "hostname": node() + _DOMAIN,
                "r": '',
                "e": False,
            }
        ).sign(self._keys.vk_Ed25519_sec_key)

    @DualUse.method()
    def get_new_system_state(self):
        old_state = self.state.system_state.copy()
        new_system_state = SystemState.read(self)
        res = self.state.event_NEW_SYSTEM_STATE(new_system_state)
        if res.error:
            click.exit("Fatal unable to handle new system state: %r" % (res,))
        if old_state == new_system_state:
            self.log.info("system state unchanged")
        else:
            # FIXME: ensure that triggers do everything necessary, and then
            # remove this full repair sync call here:
            self.sync()
            self._instruct_zeroconf()
        return res

    def _load_state(self):
        """
        Deserializes the state object from disk and returns it
        """
        self.log.debug("Loading state file")
        try:
            state = OrganizeState.from_yaml_file(self.state_file)
            self.log.debug("Loaded state with %s peers" % (len(state.peers),))
        except Exception as ex:
            self.log.info("Couldn't load state file: %r", ex)

            state = OrganizeState()

            if os.path.exists(self.state_file):
                self.log.info(
                    "Existing state file will be overwritten because it was "
                    "malformed: %r",
                    ex
                    # XXX we should probably move the malformed file aside here
                )
            self.log.debug("Created new OrganizeState")

        if state.event_log:
            self.log.info(
                "event_log contains %s entries" % (len(state.event_log),)
            )
        return state

    @property
    def state(self):
        return self._state

    @property
    def peers(self):
        return self._state.peers

    @property
    def prefs(self):
        return self._state.prefs

    @DualUse.method()
    def _write_hosts_file(self):
        """
        Write the hosts file
        """
        hosts_file: str = _ORGANIZE_HOSTS_FILE
        hosts = {
            name: list(peer.descriptor.addrs)[0]
            # XXX make this use the "best ip" logic
            for peer in self.peers.limit(enabled=True).values()
            for name in peer.enabled_names
        }
        Path(hosts_file).touch(mode=0o644)
        with click.open_file(
            hosts_file, mode='w', encoding='utf-8', atomic=True
        ) as fh:
            fh.write(
                "\n".join("%s %s" % (ip, host) for host, ip in hosts.items())
                + "\n"
            )
        chown_like_dir_if_root(hosts_file)
        return True

    @DualUse.method()
    def save(self):
        """
        Save state to disk

        (should be no-op if run from commandline in a new organize instance)
        """
        self.state.write_yaml_file(self.state_file, mode=0o600, autochown=True)
        self.log.info("vula state file updated: %i peers", len(self.peers))
        self._write_hosts_file()

    @DualUse.method()
    def verify_and_pin_peer(self, vk, hostname):
        return str(
            yamlrepr(raw(self.state.event_VERIFY_AND_PIN_PEER(vk, hostname)))
        )

    @DualUse.method(
        opts=(
            click.option('-n', '--dryrun', is_flag=True),
            click.option('-v', '--verbose', is_flag=True),
        )
    )
    def sync(self, dryrun=False, verbose=False, firstrun=False):
        """
        Sync system to the desired organized state
        """
        res = []
        res += self.sys.sync_interface(dryrun=dryrun)
        res += self.sys.sync_iprules(dryrun=dryrun)
        for peer in self.peers.values():
            self.log.debug("syncing peer %s", peer.name_and_id)
            try:
                peer_res = self.sys.sync_peer(peer.id, dryrun)
            except Exception as ex:
                peer_res = repr(ex)
            res.append(peer_res)
        res += self.sys.remove_unknown(dryrun=dryrun)
        res = list(filter(None, res))
        if res and not firstrun:
            pass
            # self.log.info("sync: %s" % (res,))
            # currently this global sync is all we have; later it will again be
            # expected to have no results and the below log line can be
            # restored:
        #            self.log.warn(
        #                "WARNING: peers %s out of sync: \n%s",
        #                ('were', 'are')[dryrun],
        #                res,
        #            )
        return res

    @DualUse.method()
    def rediscover(self):
        self.discover.listen([])
        self._instruct_zeroconf()
        return ",".join(map(str, self.state.system_state.current_ips))

    @DualUse.method()
    def release_gateway(self):
        """
        Release the current gateway
        """
        return str(yamlrepr(self.state.event_RELEASE_GATEWAY()))

    @DualUse.method()
    def bp(self):
        """
        Call pdb.set_trace()
        """
        pdb.set_trace()

    @DualUse.method()
    @click.option(
        '-m', '--monolithic', is_flag=True, help="Run in monolithic mode"
    )
    @click.option(
        '--no-dbus',
        is_flag=True,
        help="Run in monolithic mode without dbus (experimental/unsupported)",
    )
    def run(self, monolithic, no_dbus=False):
        """
        Run GLib main loop (default if no command specified)
        """

        if not no_dbus:
            system_bus = pydbus.SystemBus()
            system_bus.publish(_ORGANIZE_DBUS_NAME, self)
            main_loop = GLib.MainLoop()

        if monolithic or no_dbus:
            self.discover = Discover()
            self.discover.callbacks.append(self.process_descriptor)
            self.publish = Publish()
        else:
            self.discover = system_bus.get(
                _DISCOVER_DBUS_NAME, _DISCOVER_DBUS_PATH
            )
            self.publish = system_bus.get(
                _PUBLISH_DBUS_NAME, _PUBLISH_DBUS_PATH
            )

        # remove old listener, if there is one
        self.discover.listen([])

        self.get_new_system_state()
        self.sys.start_monitor()
        self._instruct_zeroconf()
        self.sync()

        if not no_dbus:
            self.log.info("calling GLib.MainLoop().run()")
            main_loop.run()

    def _instruct_zeroconf(self):

        descriptors = {}
        vf: int = int(time.time())
        for net, ips in self.state.system_state.current_subnets.items():
            desc = {
                k: str(v)
                for k, v in self._construct_service_descriptor(
                    ",".join(str(ip) for ip in ips), vf
                )
                ._dict()
                .items()
            }
            for ip in ips:
                descriptors[str(ip)] = desc
        current_ips = list(map(str, self.state.system_state.current_ips))
        self.log.info(
            "discovering on {ips} and publishing {pub}".format(
                ips=current_ips,
                pub='on same'
                if list(descriptors.keys()) == current_ips
                else descriptors,
            )
        )
        self.discover.listen(current_ips)
        self.publish.listen(descriptors)
        self._latest_descriptors = descriptors
        self.log.info("Current IP(s): {}".format(current_ips))
        self.log.info(
            "Current descriptors: {}".format(self._latest_descriptors)
        )

    @DualUse.method(opts=(click.argument('query', type=str),))
    def show_peer(self, query):
        """
        Returns peer description string from query for vk, hostname, or IP.
        """
        peer = self.peers.query(query)
        return (
            peer.show(self.sys.get_stats().get(str(peer.descriptor.pk)))
            if peer
            else "No peer matched query %r" % (query,)
        )

    @DualUse.method(opts=(click.argument('query', type=str),))
    def peer_descriptor(self, query):
        """
        Returns peer descriptor string from query for vk, hostname, or IP.
        """
        peer = self.peers.query(query)
        return str(peer.descriptor) if peer else ''

    def process_descriptor_string(self: Organize, descriptor: str):

        self.log.debug("about to parse descriptor: %r", descriptor)
        try:
            descriptor = Descriptor.parse(descriptor)
        except Exception as ex:
            self.log.info(
                "organize failed to parse descriptor because %r (descriptor "
                "was %r)" % (ex, descriptor)
            )
            return

        if descriptor is None:
            self.log.info(
                "organize failed to parse descriptor (descriptor was %r)"
                % (descriptor,)
            )
            return

        return self.process_descriptor(descriptor)

    def process_descriptor(self: Organize, descriptor: Descriptor):

        if not descriptor.verify_signature():
            self.log.info(
                "Discarded descriptor with invalid signature: %r"
                % (descriptor,)
            )
            return

        res = self.state.event_INCOMING_DESCRIPTOR(descriptor)

        # if res.writes:
        #    # this should happen with triggers in the event engine
        #    self.sync()
        return str(yamlrepr(res))

    def get_vk_by_name(self, hostname):
        return self.peers.with_hostname(hostname).id

    def peer_ids(self, which):
        peers = self.peers
        if which == 'disabled':
            peers = peers.limit(enabled=False)
        elif which == 'enabled':
            peers = peers.limit(enabled=True)
        else:
            assert which == 'all'
        return list(peers.keys())

    def dump_state(self, interactive, dbus_context):
        if dbus_context.is_authorized(
            'local.vula.organize1.Debug.dump_state',
            details={},
            interactive=interactive,
        ):
            return repr(yamlrepr(self.state._dict()))
        else:
            return "Forbidden"

    def test_auth(self, interactive, dbus_context):
        if dbus_context.is_authorized(
            'local.vula.organize1.Debug.test_auth',
            details={},
            interactive=interactive,
        ):
            return "OK"
        else:
            return "Forbidden"

    def set_peer(self, vk, path, value):
        res = self.state.event_USER_EDIT('SET', ['peers', vk] + path, value)
        return str(jsonrepr(res))

    def remove_peer(self, query):
        return str(yamlrepr(self.state.event_USER_REMOVE_PEER(query)))

    def peer_addr_add(self, vk, ip):
        return str(yamlrepr(self.state.event_USER_PEER_ADDR_ADD(vk, ip)))

    def peer_addr_del(self, vk, ip):
        return str(yamlrepr(self.state.event_USER_PEER_ADDR_DEL(vk, ip)))

    def show_prefs(self):
        return str(yamlrepr(self.state.prefs))

    def set_pref(self, pref, value):
        # this should call event_EDIT_PREF instead of event_USER_EDIT; this
        # as-yet unwritten event should call an action which should trigger
        # get_new_system_state to cause, eg, removal of allowed subnets
        # or interfaces to take effect immediately.
        return str(self.state.event_USER_EDIT('SET', ['prefs', pref], value))

    def add_pref(self, pref, value):
        return str(self.state.event_USER_EDIT('ADD', ['prefs', pref], value))

    def remove_pref(self, pref, value):
        return str(
            self.state.event_USER_EDIT('REMOVE', ['prefs', pref], value)
        )

    @DualUse.method()
    def eventlog(self):
        return "\n".join(
            "{event}: {actions} {writes} {triggers}".format(
                event=result.event[0],
                actions=[action[0] for action in result.actions],
                writes=[write[0] for write in result.writes],
                triggers=[trigger[0] for trigger in result.triggers],
            )
            for result in map(Result, self.state.event_log)
        )


Organize.cli.add_command(PeerCommands.cli, name='peer')
Organize.cli.add_command(PrefsCommands.cli, name='prefs')

main = Organize.cli

if __name__ == "__main__":
    main()



import time
from unittest.mock import MagicMock, patch

import vula.sys_pyroute2


class TestSys:
    def test_start_stop_monitor(self):
        mock_organize = MagicMock()
        with patch("vula.sys_pyroute2.IPRSocket") as mock_ipr, patch(
            "vula.sys_pyroute2.WgInterface"
        ):
            mock_ipr.return_value.get.return_value = [
                {"event": "NON_EXISTING"}
            ]
            sys = vula.sys_pyroute2.Sys(mock_organize)

            sys.start_monitor()
            sys.stop_monitor()
            # HACK to force monitor_thread to run
            time.sleep(0.1)

            mock_organize.log.info.assert_called_once_with(
                "Stopping netlink monitor thread"
            )
            mock_ipr.return_value.close.assert_called_once()

    def test_monitor_newneigh(self):
        mock_organize = MagicMock()
        with patch("vula.sys_pyroute2.IPRSocket") as mock_ipr, patch(
            "vula.sys_pyroute2.WgInterface"
        ):
            mock_ipr.return_value.get.return_value = [
                {"event": "RTM_NEWNEIGH"}
            ]
            sys = vula.sys_pyroute2.Sys(mock_organize)
            sys.get_new_system_state = MagicMock()
            sys._stop_monitor = True

            sys._monitor()

            sys.get_new_system_state.assert_not_called()

    def test_monitor_netlink_msg(self):
        mock_organize = MagicMock()
        with patch("vula.sys_pyroute2.IPRSocket") as mock_ipr, patch(
            "vula.sys_pyroute2.WgInterface"
        ):
            mock_ipr.return_value.get.return_value = [{"event": "RTM_NEWADDR"}]
            sys = vula.sys_pyroute2.Sys(mock_organize)
            sys.get_new_system_state = MagicMock()
            sys._stop_monitor = True

            sys._monitor()

            sys.get_new_system_state.assert_called_once()

    def test_monitor_bug(self):
        mock_organize = MagicMock()
        with patch("vula.sys_pyroute2.IPRSocket") as mock_ipr, patch(
            "vula.sys_pyroute2.WgInterface"
        ):
            mock_ipr.return_value.get.side_effect = [
                [],
                [{"event": "NON_EXISTING"}],
            ]
            sys = vula.sys_pyroute2.Sys(mock_organize)
            sys.get_new_system_state = MagicMock()
            sys._stop_monitor = True

            sys._monitor()

            sys.get_new_system_state.assert_not_called()
            assert mock_organize.log.info.call_count == 2
